#module for getting ECMWF ERA5 data time series
#note that in order to use this code you have to register as user with the CDS and save a license key file to your home directory as explained on the ERA-5 CDS website(s)
import numpy as np
import tempfile
import cdsapi
from netCDF4 import Dataset
import os
from scipy import interpolate
from datetime import datetime, timedelta

param_full_names = {
    'z': 'surface geopotential',
    'lsm':'land sea mask',
    'rsn':'snow density',
    'sd':'snow depth',
    '10u':'10m eastward wind velocity',
    '10v':'10m northward wind velocity',
    '2t':'2m temperature',
    '2d': '2m dewpoint temperature',
    'stl1':'level 1 soil temperature (1-7 cm)',
    'stl2':'level 2 soil temperature (7-28 cm)',
    'stl3':'level 3 soil temperature (28-100 cm)',
    'stl4':'level 4 soil temperature (100-289 cm)',
    'tsn':'snow temperature',
    'sf':'snowfall',
    'es':'snow evaporation',
    'smlt':'snow melt',
    'tp':'total precipitation',
    'ro':'run off'
    }

param_units = {
    'z':'m^2 s^-2',
    'lsm':'{0,1}',
    'rsn':'kg/m^3',
    'sd':'m (swe)',
    '10u':'m/s',
    '10v':'m/s',
    '2t':'K',
    '2d': 'K',
    'stl1':'K',
    'stl2':'K',
    'stl3':'K',
    'stl4':'K',
    'tsn':'K',
    'sf':'m (swe)',
    'es':'m (swe)',
    'smlt':'m (swe)',
    'tp':'m of water',
    'ro':'m of water'
    }

era5_invariant_params = ['z', 'lsm']
era5_accumulated_params = ['es', 'smlt', 'sf', 'csf', 'lsf', 'tp', 'ro']
era5_temperature_params = ['2t', '2d', 'tsn', 'stl1', 'stl2', 'stl3', 'stl4']

def get_dataset(loc, start_date, end_date, params_all, coastal=False):
    #this function gets values for all parameters of interest at hourly intervals for snow study
    # times = [
    #     '00:00', '01:00', '02:00',
    #     '03:00', '04:00', '05:00',
    #     '06:00', '07:00', '08:00',
    #     '09:00', '10:00', '11:00',
    #     '12:00', '13:00', '14:00',
    #     '15:00', '16:00', '17:00',
    #     '18:00', '19:00', '20:00',
    #     '21:00', '22:00', '23:00',
    # ]
    times=['00:00']


    params_land = [param for param in params_all if param not in era5_invariant_params]
    params_invariant = [param for param in params_all if param in era5_invariant_params]

    #build area as [N, W, S, E] from [lon, lat] with special handling for coast areas
    if coastal:
        raise ValueError('coastal flag not implemented')
        #special handling with the coastal flag on
        #loc = _get_inland_loc(loc)
    area = [loc[1], loc[0], loc[1], loc[0]]

    return_dict = {}
    return_dict['start_date'] = start_date
    return_dict['end_date'] =end_date
    #return_dict['dt'] = 1./24
    return_dict['dt'] = 1.
    client = cdsapi.Client()
    if len(params_land) != 0:
        #build dates list
        start_date_ =  datetime.strptime(start_date, '%Y%m%d')
        end_date_ = datetime.strptime(end_date, '%Y%m%d')
        ndays = int((end_date_ - start_date_).total_seconds()/86400. + 0.5) + 1
        del_1d = timedelta(days=1)
        dates = [start_date_ + ii*del_1d for ii in np.arange(ndays)]
        dates = [date_.strftime('%Y%m%d') for date_ in dates]

        #loop over each param to avoid exceeding query item limit of 100000
        #get temp file to receive query result
        _, target = tempfile.mkstemp(suffix = '.era.nc')

        #execute query
        print("\nRetrieving ERA5 land params to %s \n" % target)
        client.retrieve(
            'reanalysis-era5-land',
            {
                'format': 'netcdf',
                'time': times,
                'date': dates,
                'variable': params_land,
                'area': area
            },
            target)

        #read resulting netcdf4 format file and extract times and param vectors
        rootgrp = Dataset(target, 'r', format='NETCDF4')
        lons = rootgrp['longitude'][:]
        lats = rootgrp['latitude'][:]
        time = rootgrp['time'][:]
        t0 = datetime(1900,1,1)
        del_1h = timedelta(hours=1)
        return_dict['times'] = [t0 + time_*del_1h for time_ in time]
        if (len(lons) > 1) or (len(lats) > 1):
            raise ValueError('Unexpected number of ERA5 grid points returned: expecting 1')

        # the netcdf variable name sometimes differs from the param code used for the query so this dict provides these special return codes
        _param_override_return_codes_era5 = {
            '10u': 'u10',  # 10m eastward wind (m/s)
            '10v': 'v10',  # 10m northward wind (m/s)
            '2t': 't2m',  # 2m temperature (K)
            '2d': 'd2m',  # 2m dewpoint temperature (K)
        }
        for param in params_land:
            if param in _param_override_return_codes_era5:
                return_dict[param] = np.squeeze(rootgrp[_param_override_return_codes_era5[param]][:])
            else:
                return_dict[param] = np.squeeze(rootgrp[param][:])

        #close, delete temp file
        rootgrp.close()
        os.remove(target)

        #special post-treatment
        for ii, param in enumerate(params_land):
            if param in era5_temperature_params:
                return_dict[param] -= 273.15
            if param in era5_accumulated_params:
                nstep = len(times)
                if nstep > 1:
                    # returned values are accumulations since start of the day but we want accumulation over each step
                    idx = np.arange(len(return_dict['times']))
                    orig = np.copy(return_dict[param])
                    return_dict[param][0] = 0.
                    return_dict[param][1:] = orig[1:] - orig[0:-1]
                    is_step_1 = (idx % nstep) == 1
                    return_dict[param][is_step_1] = orig[is_step_1]

    #get invariant params (requires seperate query because they are in a separate dataset
    if len(params_invariant) != 0:
        _, target = tempfile.mkstemp(suffix = '.era.nc')
        print("\nRetrieving ERA5 invariant params to %s \n" % target)
        client.retrieve(
            'reanalysis-era5-single-levels-monthly-means',
            {
                'format': 'netcdf',
                'variable': params_invariant,
                'year': '2019',
                'month': '01',
                'time': '00:00',
                'area': area,
            },
            target)
        rootgrp = Dataset(target, 'r', format='NETCDF4')
        lons = rootgrp['longitude'][:]
        lats = rootgrp['latitude'][:]
        if (len(lons) > 1) or (len(lats) > 1):
            raise ValueError('Unexpected number of ERA5 grid points returned: expecting 1')
        for param in params_invariant:
            return_dict[param] = np.squeeze(rootgrp[param][:])[0]
        # close, delete temp file
        rootgrp.close()
        os.remove(target)
        x=1
    return return_dict

def oversample_dataset(dict, new_dt):
    return_dict = {}
    return_dict['dt'] = new_dt
    dt_ratio = dict['dt']/return_dict['dt']
    return_dict['times'] = [dict['times'][0] + datetime.timedelta(return_dict['dt'] * ii) for ii in
                            np.arange(int(dt_ratio*len(dict['times'])))]
    times_orig = np.arange(0, len(dict['times']))*dict['dt']
    times_new = np.arange(0, len(return_dict['times']))*return_dict['dt']
    return_dict['start_date'] = dict['start_date']
    return_dict['end_date'] =dict['end_date']

    #now run through all keys in dict and interpolate them onto new time grid
    for key in dict:
        if key in era5_invariant_params:
            return_dict[key] = dict[key]

        elif key == '10d':
            raise ValueError('wind direction should be computed after 10u/10v components are resampled')

        elif key not in era5_invariant_params:
            return_dict[key] = interpolate.interp1d(times_orig, dict[key], assume_sorted=True, fill_value='extrapolate')(times_new)
            #scale accumulation codes by inverse of oversampling ratio
            if key in era5_accumulated_params:
                return_dict[key] /= dt_ratio

    return return_dict

def get_derived_codes(dict):
    #adds derived codes not provided with era5

    dict_new = {}
    # relative humidity
    def es(t):
        # implements the  August-Roche-Magnus formula for saturation vapour pressure
        # t: temperature in Celcius
        # see Lawrence, Mark.  American Metr. Soc 2005 eq 6
        return 6.1094 * np.exp(17.625 * t / (t + 243.04))

    if ('2d' in dict) and ('2t' in dict):
        dict_new['rh'] = 100 * es(dict['2d']) / es(dict['2t'])

    #wind
    if ('10u' in dict) and ('10v' in dict):
        dict_new['10m'] = (dict['10u']**2 + dict['10v']**2)**0.5
        dict_new['10d'] = (((-1*np.arctan2(-1*dict['10v'], -1*dict['10u'])*180/np.pi)+90.0 +360)% 360)

    return dict_new

def test():
    era5_codes = ['rsn', 'sd', '10u', '10v', '2t', '2d', 'tsn', 'stl1', 'stl2', 'stl3', 'stl4', 'es', 'smlt', 'sf', 'tp', 'ro']
    get_dataset([-81.11, 50.22], '20181101', '20181102', ['lsm', 'z', 'sf', '10v', '10u', '2d', '2t'])


if __name__ == '__main__':
    test()
    pass