__author__ = "Jayson Eppler"
__copyright__ = "Copyright 2022, Jayson Eppler"
__credits__ = ["Jayson Eppler"]
__license__ = "GPL"
__version__ = "1.0.1"
__maintainer__ = "Jayson Eppler"
__email__ = "jaysone@sfu.ca"

# Contains code used to implement Snow Water Equivalent (SWE) mapping from repeat pass InSAR phase
# maps based on spatial correlation of dry-snow refractive phase and topographic slope.
#
# This code was used to produce all results and figures for the paper submitted to The Cryosphere Disscusions:
# the paper submitted to The Cryosphere Disscus:
# [Eppler, et al,. accepted for publication](https://doi.org/10.5194/tc-2021-359)

from matplotlib import pyplot as plt, rcParams, cm
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import subprocess
import sys
import os
import glob
from os.path import exists
from datetime import datetime, timedelta
from osgeo import gdal
from scipy import constants
import scipy.optimize as opt
from scipy.ndimage import uniform_filter
import scipy.ndimage.morphology as morph

from sarlab.tools.conversion import utm_to_geo
from sarlab.gammax import DEM_Par, readBin, cmap_gamma, MLI_Par, writeBin, run, read_ras, ensureDir, exec, SLC_Par, DEM, SLC_stack
from sarlab import multiprocess_gx as mp
from sarlab.tools.least_squares import lsq2_fit, poly2
from sarlab.tools import render_png
from sarlab.met.env_canada import parse_ec_dir

import .snow as snow
from .snow_survey_measurements import t_sa_2018, t_sb_2018, t_sc_2018, t_ensom_2018, transects_enr

#setup for paper figures
fig_dir = './paper_repeat_pass_snow/figs/'
if not os.path.exists(fig_dir):
    os.makedirs(fig_dir)

fig_width = 3.5
fig_width_full = 7.16

fig_params = {
        'axes.titlesize':8,
        'axes.labelsize': 8,
        'font.size': 8,
        'legend.fontsize': 8,
        'legend.frameon': False,
        'xtick.labelsize': 8,
        'ytick.labelsize': 8,
        'text.usetex': False,
        'figure.dpi': 96,
        'figure.autolayout': False,
        'image.interpolation': 'nearest',
        'savefig.dpi': 600,
        'mathtext.default': 'regular'
   }
rcParams.update(fig_params)
plt_formats = ['k-', 'k--', 'k-.', 'k:']
cmap_gamma_ma = cmap_gamma()
cmap_gamma_ma.set_bad('k', 1.0)
cmap_viridis_ma = plt.cm.viridis
cmap_viridis_ma.set_bad('k', 1.0)

cmap_rdbu = cm.get_cmap('RdBu')
cmap_rdbu.set_bad(color='grey')
cmap_viridis = cm.get_cmap('viridis')
cmap_viridis.set_bad(color='grey')

master_SLA27D = '20160811'
ingest_cfg = {'polarizations': None}
stack_SLA27D = SLC_stack(dirname= './inuvik/RS2_SLA27_D/full_scene/',name='inuvik_RS2_SLA27_D', master=master_SLA27D, looks_hr=(1,4), looks_lr=(11,44), multiprocess=True, rdcdem_refine=True, skipmode='None', ingest_cfg=ingest_cfg)

#gives 1000m x 1000m (ground rng x az) for SLA27D at 3x12 looks
win1000 = np.asarray((188, 204))
cfg_default = {'good_crop': [88, 96, 3912, 1991], 'win':win1000//2, 'stack':stack_SLA27D, 'looks':(3,12), 'looks_str':'3_12', 'swe_range':(-0.05, 0.08), 'swe_del':0.002, 'swe_est_dir': stack_SLA27D._dir+'swe_est_3_12/', 'rho':0.3}


def exec(cmd, input=None, print_cmd=True, can_throw_exception=True):
    #executes a shell command with optional input
    #outputs command to terminal and checks for return code
    if isinstance(cmd, str):
        args = cmd.split()
    else:
        args = cmd

    cmd_string = '\'' + args[0]
    for arg in args[1:]:
        cmd_string = cmd_string + ' ' + arg
    if print_cmd:
        print(cmd_string + '\' ...', end='')
    sys.stdout.flush()
    runOutput = subprocess.run(args, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True, input=input)
    if runOutput.returncode == 0:
        if print_cmd:
            print('completed.')
        sys.stdout.flush()
    else:
        print('\nError encountered in: ' + args[0])
        print(runOutput.stdout)
        print(runOutput.stderr)
        sys.stdout.flush()
        if can_throw_exception:
            raise ValueError('Non-zero return code from gammax.exec')
    return runOutput

def snow_real_permittivity(rho):
    #from Leinss 2015 eq(3)
    #rho is snow density in g/cm^3
    #no boundary at 0.4
    if rho > 1.:
        raise ValueError('rho>1 is not allowed!')

    a1 = 1.5995  # cm^3/g
    a3 = 1.861  # cm^9/g^3
    eps = 1+a1*rho+a3*rho**3

    return eps

def get_inuvik_RS2_SLA27D_era5():
    inuvik_RS2_SLA27D_scene_centre = [-133.6444803, 68.3595694]
    return snow.get_era5_snow_study_series(inuvik_RS2_SLA27D_scene_centre, '20110701', '20210131')


#just changed for continuity of eps(rho) across rho=04 boundary
def dphi_dswe(u, inc, rho, lamb):
    if True:
        eps = np.asarray([snow_real_permittivity(rho_) for rho_ in rho])
    else:
        eps = snow_real_permittivity(rho)

    xi = 4*np.pi/lamb/rho*(np.sqrt(eps-1+np.cos(inc)**2)-np.cos(inc)) * np.cos(u)
    return xi

def dphi_dheave(inc, lamb):
    _d = -4*np.pi/lamb*np.cos(inc)
    return _d

def effect_of_wrong_density():
    stack = stack_SLA27D
    #generate plot of fractional swe estimation error caused by wrong density
    del_rho = 0.001
    rho_actual = np.arange(0.1, 0.5, del_rho)
    rho_est = np.arange(0.1, 0.5, del_rho)
    par = SLC_Par(stack._dir + 'rslc/master.rslc.par')
    lamb = constants.c/par['radar_frequency']
    inc = np.deg2rad(par['incidence_angle'])

    rho_actual_grid, rho_est_grid = np.meshgrid(rho_actual, rho_est)
    frac_bias = np.reshape(dphi_dswe(0., inc, rho_actual_grid.flatten(), lamb)/dphi_dswe(0., inc, rho_est_grid.flatten(), lamb)-1, rho_actual_grid.shape)
    plt.figure(figsize=(fig_width, 1.7))
    plt.subplot(1,2,1)
    ax = plt.gca()
    levels = np.asarray([-6,-5,-4,-3,-2,-1,0,1,2,3,4,5,6])
    def fmt(x):
        s = f"{x:.1f}"
        if s.endswith("0"):
            s = f"{x:.0f}"
        return rf" {s}\% " if plt.rcParams["text.usetex"] else f" {s}% "
    plt.rcParams['contour.negative_linestyle'] = 'solid'
    CS = ax.contourf(rho_actual_grid, rho_est_grid, frac_bias * 100., levels, cmap='RdBu', vmin=-10, vmax=10, aspect=1.)
    CS = ax.contour(rho_actual_grid, rho_est_grid, frac_bias * 100., levels, colors='k', aspect=1, linewidths=1)
    plt.xlabel(r'True $\rho$')
    plt.ylabel(r'Assumed $\rho$')
    ax.clabel(CS, inline=True, fmt=fmt, fontsize=6, manual=True)
    plt.xticks([0.1, 0.2, 0.3, 0.4, 0.5])
    plt.yticks([0.1, 0.2, 0.3, 0.4, 0.5])

    #now plot apparent SWE change caused by density changes
    rho_1 = np.arange(0.1, 0.5, del_rho)
    rho_2 = np.arange(0.1, 0.5, del_rho)
    rho_1_grid, rho_2_grid = np.meshgrid(rho_1, rho_2)
    xi_1 = np.reshape(dphi_dswe(0., inc, rho_1_grid.flatten(),lamb), rho_1_grid.shape)
    xi_2= np.reshape(dphi_dswe(0., inc, rho_2_grid.flatten(), lamb), rho_2_grid.shape)
    apparent_swe_change = (xi_2 - xi_1)/((xi_2 + xi_1)/2)
    plt.subplot(1,2,2)
    ax = plt.gca()
    levels = np.asarray([-6,-5,-4,-3,-2,-1,0,1,2,3,4,5,6])

    plt.rcParams['contour.negative_linestyle'] = 'solid'
    CS = ax.contourf(rho_1_grid, rho_2_grid, apparent_swe_change * 100., levels, cmap='RdBu', vmin=-10, vmax=10, aspect=1.)
    CS = ax.contour(rho_1_grid, rho_2_grid, apparent_swe_change * 100., levels, colors='k', aspect=1, linewidths=1)
    plt.xlabel(r'Initial $\rho$')
    plt.ylabel(r'Final $\rho$')
    ax.clabel(CS, inline=True, fmt=fmt, fontsize=6, manual=True)
    plt.xticks([0.1, 0.2, 0.3, 0.4, 0.5])
    plt.yticks([0.1, 0.2, 0.3, 0.4, 0.5])
    plt.tight_layout()
    plt.savefig(fig_dir + 'density_error_effects.png')

#generates a water mask based on temporal min amplitude image
def mk_water_mask(indir, outdir, dates, iterations=1, SLA=False):
    #hand-picked to speed up and improve statistics
    fnames = [indir + date + '.rmli' for date in dates]
    par = MLI_Par(indir+'ave.rmli.par')
    rmli_min = np.zeros(par.dim)
    N=0
    for fname in fnames:
        print('reading ', fname, ' ...')
        rmli = readBin(fname, par.dim, 'float32')
        rmli[rmli == 0] = 1000.
        rmli = uniform_filter(rmli, 5)
        if N==0:
            rmli_min = rmli
        else:
            rmli_min = np.minimum(rmli_min, rmli)
        N+=1
    rmli_min[rmli_min <= 0] = np.min(rmli_min[rmli_min > 0])
    writeBin(outdir + 'min.rmli', rmli_min)
    rmli_min = readBin(outdir + 'min.rmli', par.dim, 'float32')
    if SLA:
        #spotlight so special treatment for variable noise floor
        thresh_dB =10*np.log10(np.min(rmli_min[0:4000,:], axis=0))+3
        plt.figure()
        plt.plot(thresh_dB)
        plt.ylim((-30, None))
    else:
        thresh_dB = -21.

    rmli_min_db = 10*np.log10(rmli_min)
    water_mask_init = np.zeros(par.dim)
    for ii in np.arange(par.dim[0]):
        water_mask_init[ii, rmli_min_db[ii,:] < thresh_dB] = 1.
    water_mask = morph.binary_closing(morph.binary_opening(water_mask_init, iterations=iterations), iterations=iterations)
    writeBin(outdir + 'water_mask', water_mask.astype('float32'))
    plt.figure()
    ax1 = plt.subplot(3,1,1)
    plt.imshow(rmli_min_db.T, vmin=-30, vmax=-10)
    plt.subplot(3,1, 2, sharex=ax1, sharey=ax1)
    plt.imshow(water_mask_init.T, vmin=0, vmax=1.)
    plt.subplot(3,1, 3, sharex=ax1, sharey=ax1)
    plt.imshow(water_mask.T, vmin=0, vmax=1.)
    render_png((water_mask.T).astype('float'),outdir + 'water_mask_autogen.png', vmin=0, vmax=1.)

    plt.show()

def render_terrain_data(cfg):
    print('Rendering terrain maps.')
    stack = cfg['stack']
    looks_str = cfg['looks_str']

    dir = stack._dir + 'dem_' + looks_str + '/'
    dir = dir + 'topo/'
    par = MLI_Par(stack._dir + 'rmli_'+looks_str+ '/rmli_'+looks_str+'.ave.par')
    lamb = constants.c / par['radar_frequency']

    basins = readBin(dir+'basins_all.rdc', par.dim, 'float32')
    render_png(basins.T, dir + 'basins_all.rdc.png', vmin=0., vmax=4.)

    tpi = readBin(dir+'tpi.rdc', par.dim, 'float32')
    #render_png(tpi.T, dir + 'tpi.rdc.png', cmap='terrain_r', vmin=5., vmax=12.)
    render_png(tpi.T, dir + 'tpi.rdc.png', vmin=5., vmax=12.)

    slope = readBin(dir+'slope.rdc', par.dim, 'float32')
    render_png(slope.T, dir + 'slope.rdc.png', vmin=0., vmax=15.)

    aspect = readBin(dir+'aspect.rdc', par.dim, 'float32')
    render_png(aspect.T, dir + 'aspect.rdc.png', cmap=cmap_gamma(), vmin=-np.pi,vmax=np.pi)  # north is zero, east is pi/2, west is -pi/2

    inc = readBin(dir+'inc.rdc', par.dim, 'float32')
    render_png(np.degrees(inc).T, dir + 'inc.rdc.png', vmin=18., vmax=35.)
    #render_png(np.degrees(inc).T, dir + 'inc.rdc.png', vmin=40, vmax=60)

    _dphi_dswe = readBin(dir+'dphi_dswe.rdc', par.dim, 'float32')
    render_png(_dphi_dswe.T / 2 / np.pi, dir + 'dphi_dswe.rdc.png', vmin=29., vmax=34.)
    #render_png(_dphi_dswe.T / 2 / np.pi, dir + 'dphi_dswe.rdc.png', vmin=35., vmax=45.)

    dphi_ddownslope = readBin(dir + 'dphi_ddownslope.rdc', par.dim, 'float32')
    render_png(dphi_ddownslope.T, dir + 'dphi_ddownslope.rdc.png', cmap = 'RdBu', vmin=-4*np.pi/lamb, vmax=4*np.pi/lamb)

    dphi_ddownslope_slope_scaled = readBin(dir + 'dphi_ddownslope_slope_scaled.rdc', par.dim, 'float32')
    # lim_min = np.percentile(dphi_ddownslope_slope_scaled.flatten(),1)
    # lim_max = np.percentile(dphi_ddownslope_slope_scaled.flatten(),99)
    # lim = np.max(np.abs(np.asarray((-lim_min, lim_max))))
    lim=16
    render_png(dphi_ddownslope_slope_scaled.T, dir + 'dphi_ddownslope_slope_scaled.rdc.png', cmap = 'RdBu', vmin=-lim, vmax=lim)

    _dphi_dheave = readBin(dir+'dphi_dheave.rdc', par.dim, 'float32')
    render_png(_dphi_dheave.T / 2 / np.pi, dir + 'dphi_dheave.rdc.png')


def get_band(fname,crop=None):
    ds = gdal.Open(fname)
    band_obj = ds.GetRasterBand(1)
    arr = band_obj.ReadAsArray()
    if crop is not None:
        arr = arr[crop[0]:crop[0]+crop[2], crop[1]:crop[1]+crop[3]]
    return arr

def nan_fill(arr_, win = (3,3)):
    print('nan_fill...')
    arr = np.copy(arr_)
    bad_mask = ~np.isfinite(arr)
    num_bad = np.sum(bad_mask)
    ii=0
    while(num_bad > 0):
        print('iteration:', ii+1, ', found', num_bad, 'bad values.')
        arr[bad_mask] = 0
        norm_filt = uniform_filter((~bad_mask).astype(float), win)
        norm_filt[norm_filt < 1/win[0]/win[1]]=0
        arr_filt = uniform_filter(arr, win)/norm_filt
        arr[bad_mask] = arr_filt[bad_mask]
        bad_mask = ~np.isfinite(arr)
        num_bad = np.sum(bad_mask)
        ii+=1

    return arr

def make_terrain_maps(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    #create terrain related maps from source DEM
    rho = cfg['rho'] #normalized snow density
    init_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/dem_cropped_gaussrad3.tif'
    utm_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/dem_cropped_utm.tif'
    slope_utm_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/slope_utm.tif'
    aspect_utm_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/aspect_utm.tif'
    aspect_n_utm_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/aspect_n_utm.tif'
    aspect_e_utm_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/aspect_e_utm.tif'
    tpi_utm_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/tpi_raw.tif'
    basins_utm_fn = '/local-scratch/users/jaysone/projects_active/inuvik/dems/tdx/topo_indices/all_basins.tif'

    dem_dir = stack._dir + 'dem_' + looks_str + '/'
    rmli_dir = stack._dir + 'rmli_' + looks_str + '/'
    out_dir = dem_dir + 'topo/'
    par = MLI_Par(stack._dir + 'rmli_'+looks_str+ '/rmli_'+looks_str+'.ave.par')

    dem_par = DEM_Par(dem_dir+'seg.dem_par')
    gc_map = dem_dir+'gc_fine'
    dem_fn = dem_dir+'seg.dem'
    lamb = constants.c / par['radar_frequency']

    print('Converting basins map to rdc.')
    par.geotiff2rdc(basins_utm_fn, out_dir + 'basins_all.rdc', dem_par, gc_map, interp_mode=1)

    print('Converting source DEM to UTM.')
    warp_to_utm_args = ['gdalwarp', '--config', 'GDAL_CACHEMAX', '500', '-wm', '500', '-t_srs','EPSG:32608','-r', 'cubicspline', '-of', 'GTiff', '-overwrite'] + [init_fn] + [utm_fn]
    result = _exec(warp_to_utm_args)

    print('Making slope map.')
    slope_args = ['gdaldem', 'slope', '-of', 'GTiff', '-b', '1', '-s', '1.0'] + [utm_fn] + [slope_utm_fn]
    _exec(slope_args)
    par.geotiff2rdc(slope_utm_fn, out_dir + 'slope_init.rdc', dem_par, gc_map)
    slope = np.deg2rad(readBin(out_dir+'slope_init.rdc', par.dim, 'float32'))

    #set water areas to have zero slope
    water_mask = get_water_mask(cfg)
    print(np.sum(water_mask))
    slope[water_mask] = 0.

    #fill DEM blunder areas
    #blunders = read_ras(dem_dir + 'blunders_'+looks_str+'.ras')[0].T == 0
    #blunders = morph.binary_dilation(blunders, iterations=5)

    #slope[blunders] = np.nan
    slope = nan_fill(slope)
    writeBin(out_dir+'slope.rdc', np.rad2deg(slope))

    print('Making aspect map.')
    aspect_args = ['gdaldem', 'aspect', '-of', 'GTiff', '-b', '1', '-alg', 'ZevenbergenThorne'] + [utm_fn] + [aspect_utm_fn]
    result = _exec(aspect_args)
    #read in GDAL computed aspect angle and convert to north and east basis components (required for warping)
    aspect = get_band(aspect_utm_fn)
    aspect_n = np.cos(np.deg2rad(aspect))
    aspect_e = np.sin(np.deg2rad(aspect))
    in_raster = gdal.Open(aspect_utm_fn)
    driver = gdal.GetDriverByName('GTiff')
    out_raster = driver.Create(aspect_e_utm_fn, aspect.shape[1], aspect.shape[0], 1, gdal.GDT_Float32)
    out_raster.SetGeoTransform(in_raster.GetGeoTransform())
    out_band = out_raster.GetRasterBand(1)
    out_band.WriteArray(aspect_e)
    out_raster.SetProjection(in_raster.GetProjection())
    out_band.FlushCache()
    out_raster = None
    out_band = None
    in_raster = gdal.Open(aspect_utm_fn)
    driver = gdal.GetDriverByName('GTiff')
    out_raster = driver.Create(aspect_n_utm_fn, aspect.shape[1], aspect.shape[0], 1, gdal.GDT_Float32)
    out_raster.SetGeoTransform(in_raster.GetGeoTransform())
    out_band = out_raster.GetRasterBand(1)
    out_band.WriteArray(aspect_n)
    out_raster.SetProjection(in_raster.GetProjection())
    out_band.FlushCache()
    out_raster = None
    out_band = None
    par.geotiff2rdc(aspect_n_utm_fn, out_dir+'aspect_n.rdc', dem_par, gc_map)
    par.geotiff2rdc(aspect_e_utm_fn, out_dir+'aspect_e.rdc', dem_par, gc_map)
    aspect_n = readBin(out_dir + 'aspect_n.rdc', par.dim, 'float32')
    aspect_e = readBin(out_dir + 'aspect_e.rdc', par.dim, 'float32')

    #fill blunders in aspect components
    #aspect_n[blunders] = np.nan
    aspect_n = nan_fill(aspect_n)
    #aspect_e[blunders] = np.nan
    aspect_e = nan_fill(aspect_e)

    #compute and write aspect
    aspect = np.angle(aspect_n+1j*aspect_e)
    aspect[water_mask] = 0.
    writeBin(out_dir + 'aspect.rdc', aspect)
    aspect = readBin(out_dir+'aspect.rdc', par.dim, 'float32')

    print('Computing SAR look vectors.')
    run('look_vector', par, None, dem_par, dem_fn, dem_dir+'lv_theta', dem_dir+'lv_phi')
    run('geocode', gc_map, dem_dir+'lv_theta', dem_par.dim[0], dem_dir+'lv_theta.rdc', par.dim[0], par.dim[1], 0, 0, None, None, 2, 32, None)
    run('geocode', gc_map, dem_dir + 'lv_phi', dem_par.dim[0], dem_dir + 'lv_phi.rdc', par.dim[0], par.dim[1], 0, 0, None, None, 2, 32, None)
    lv_theta = readBin(dem_dir+'lv_theta.rdc', par.dim, 'float32')
    lv_phi = readBin(dem_dir+'lv_phi.rdc', par.dim, 'float32')

    print('Computing local incidence angle.')
    # #compute terrain slope normal vector in local ENU system
    slope_cos = np.cos(slope)
    slope_sin = np.sin(slope)
    aspect_cos = np.cos(aspect)
    aspect_sin = np.sin(aspect)
    t_u = slope_cos
    t_n = slope_sin*aspect_cos
    t_e = slope_sin*aspect_sin
    # #compute look vector in local ENU system
    lv_theta_cos = np.cos(lv_theta)
    l_u = np.sin(lv_theta)
    l_n = lv_theta_cos*np.sin(lv_phi)
    l_e = lv_theta_cos*np.cos(lv_phi)

    # #compute incidence angle
    inc = np.arccos(l_u*t_u + l_n*t_n + l_e*t_e)
    writeBin(out_dir + 'inc.rdc', inc)
    inc = readBin(out_dir + 'inc.rdc', par.dim, 'float32')

    print('Computing heave sensitivity.')
    _dphi_dheave = dphi_dheave(inc, lamb)
    writeBin(out_dir + 'dphi_dheave.rdc', _dphi_dheave)

    print('Computing topo sensitivity of dry snow phase.')
    _dphi_dswe_init = dphi_dswe(slope, inc, rho, lamb)

    #_dphi_dswe_init[blunders] = np.nan
    _dphi_dswe = nan_fill(_dphi_dswe_init)

    writeBin(out_dir + 'dphi_dswe.rdc', _dphi_dswe)

    print('Computing down-slope sensitivity.')
    #compute down-slope direction vector
    d_u = slope_sin
    d_n = slope_cos*aspect_cos
    d_e = slope_cos*aspect_sin

    #compute down-slope projection to los
    dphi_ddownslope = 4*np.pi/lamb*(-1*(l_u*d_u + l_n*d_n + l_e*d_e))
    writeBin(out_dir + 'dphi_ddownslope.rdc', dphi_ddownslope)

    #compute down-slope projection to los
    dphi_ddownslope_slope_scaled = 4*np.pi/lamb*(-1*(l_u*d_u + l_n*d_n + l_e*d_e))*np.tan(slope)
    writeBin(out_dir + 'dphi_ddownslope_slope_scaled.rdc', dphi_ddownslope_slope_scaled)

    #remap tpi map
    print('Making tpi map.')
    par.geotiff2rdc(tpi_utm_fn, out_dir + 'tpi.rdc', dem_par, gc_map)

def phase_coh(phi, win, normalize = True):
    if normalize:
        #this should be set to false only for data that has already been normalized as a time saver
        phi/=np.abs(phi)
    phi_filt = np.empty_like(phi)
    phi[~np.isfinite(phi)]=0
    phi_filt.real = uniform_filter(phi.real, win)
    phi_filt.imag = uniform_filter(phi.imag, win)
    coh = np.abs(phi_filt)
    return coh

def est_swe(phi, xi, swe_range=(-0.1, 0.1), swe_del = 0.002, win=(64,64), exclude_mask=None, weighted=False):
    #estimate swe map
    n_interp=3
    n_interp_rad = (n_interp - 1) // 2
    valid_thresh = 0 #need at least this fraction of valid points under esimation window to be considered a valid estimation
    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del
    n_swes = len(swes)
    dim = phi.shape
    if not weighted:
        #no weights so just normalize each phasor to unit length
        phi = phi / np.abs(phi)
    phi[np.isnan(phi)] = 0.
    if exclude_mask is not None:
        phi[exclude_mask] = 0.
        exclude_mask = None
    valid_mask = (phi != 0)
    abs_filt = uniform_filter(np.abs(phi), win)
    valid_mask_final = valid_mask
    valid_mask=None
    interp_points = np.zeros((n_interp, dim[0], dim[1]), dtype='float32') - 1.  # n_interp points in vicinity of peak
    idx_max = np.zeros(dim, dtype=int) - 1
    buffer = np.zeros((n_interp, dim[0], dim[1]), dtype='float32')  # last n_interp points
    for ii in np.arange(n_swes):
        # print(ii+1, 'of', n_swes)
        phi_demod_ii = phi * np.exp(-1j * swes[ii] * xi)
        coh_ii = phase_coh(phi_demod_ii, win, normalize=False) / abs_filt
        if ii < n_interp:
            # initial phase, just populate buffer
            buffer[ii, :, :] = coh_ii
        if ii >= n_interp:
            # later phase, shift buffer, then add latest entry to end
            buffer[0:-1, :, :] = buffer[1:, :, :]
            buffer[-1, :, :] = coh_ii
            # find targets with new max coh
            update_mask = buffer[n_interp_rad, :, :] > interp_points[n_interp_rad, :, :]
            # print(np.sum(update_mask), 'new peak coherences found.')
            # update interp_points for those targets with new max coh
            interp_points[:, update_mask] = buffer[:, update_mask]
            idx_max[update_mask] = ii - n_interp_rad
    buffer = None
    phi=None
    xi=None
    phi_demod_ii=None
    coh_ii = None
    update_mask=None
    valid_mask_filt=None
    swe_est = np.reshape(swes[idx_max], dim)
    idx_max=None

    # quadratic components (hardcoded for n_interp = 3)
    a = interp_points[1, :, :]
    b = -0.5 * interp_points[0, :, :] + 0.5 * interp_points[2, :, :]
    c = 0.5 * interp_points[0, :, :] - 1.0 * interp_points[1, :, :] + 0.5 * interp_points[2, :, :]
    interp_points=None
    dx = -b / 2 / c
    # handle cases where center point is not the highest
    dx[dx < -1] = -1
    dx[dx > 1] = 1.

    swe_est = swe_est + dx * swe_del
    coh_max = a + b * dx + c * dx ** 2
    a = None; b=None; c=None; dx=None

    swe_est[~valid_mask_final] = np.nan
    coh_max[~valid_mask_final] = np.nan

    return swe_est, coh_max

def est_swe_from_file(ifg_fname, par, crop, xi, out_dir, swe_range=(-0.1, 0.1), swe_render_lim = None, swe_del = 0.002, win=(64,64), exclude_mask=None):
    #estimate swe from an igram on disk and write + render results to disk
    print('Estimating SWE for:', ifg_fname)
    outname = out_dir + os.path.basename(ifg_fname)
    phi = readBin(ifg_fname, par.dim, 'complex64', crop=crop)
    swe_est_refined, coh_max_refined = est_swe(phi, xi, swe_range=swe_range, swe_del=swe_del, win=win, exclude_mask=exclude_mask)
    writeBin(outname + '.swe', swe_est_refined)
    writeBin(outname + '.swe.coh', coh_max_refined)

def est_swe_from_file_render(ifg_fname, par, crop, xi, out_dir, swe_range=(-0.1, 0.1), swe_render_lim = None, swe_del = 0.002, win=(64,64), exclude_mask=None):
    print('Rendering SWE estimates for:', ifg_fname)
    if swe_render_lim is None:
        swe_render_lim = np.max(np.abs(swe_range))
    outname = out_dir + os.path.basename(ifg_fname)
    dim = par.dim
    if crop is not None:
        dim = crop[2:]
    swe_est_refined = readBin(outname + '.swe', dim, 'float32',crop=crop)
    render_png(swe_est_refined.T, outname + '.swe.png', cmap='RdBu', vmin=-swe_render_lim,vmax=swe_render_lim)
    coh_max_refined = readBin(outname + '.swe.coh', dim, 'float32',crop=crop)
    render_png(coh_max_refined.T, outname + '.swe.coh.png', vmin=0, vmax=1)

def swe_est_clicky_func(pt_xy, cfg, phi_full, xi_full, exclude_mask_full, plots_for_paper=False):
    swe_range = (-0.05, 0.08)
    swe_range = (-0.5, 0.8)
    swe_del = 0.002
    win = win1000
    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del
    swe_render_lim = np.max(np.abs(swe_range))
    n_swes = len(swes)

    win = np.asarray(win)
    cmin = pt_xy - win // 2
    cmax = cmin + win

    # extract local patch
    phi = phi_full[cmin[0]:cmax[0], cmin[1]:cmax[1]]
    xi = xi_full[cmin[0]:cmax[0], cmin[1]:cmax[1]]
    exclude_mask = exclude_mask_full[cmin[0]:cmax[0], cmin[1]:cmax[1]]
    phi[exclude_mask]=0
    n_interp = 3
    n_interp_rad = (n_interp - 1) // 2

    interp_points = np.zeros((n_interp), dtype='float32') - 1.  # n_interp points in vicinity of peak
    idx_max =  -1
    buffer = np.zeros((n_interp), dtype='float32')  # last n_interp points
    coh = np.zeros(n_swes)
    for ii in np.arange(n_swes):
        phi_demod_ii = phi * np.exp(-1j * swes[ii] * xi)
        coh[ii] = np.abs(np.mean(phi_demod_ii))/np.mean((~exclude_mask).astype(float))
        if ii < n_interp:
            # initial phase, just populate buffer
            buffer[ii] = coh[ii]
        if ii >= n_interp:
            # later phase, shift buffer, then add latest entry to end
            buffer[0:-1] = buffer[1:]
            buffer[-1] = coh[ii]
            # find targets with new max coh
            update_mask = buffer[n_interp_rad] > interp_points[n_interp_rad]
            # print(np.sum(update_mask), 'new peak coherences found.')
            # update interp_points for those targets with new max coh
            if update_mask:
                interp_points[:] = buffer[:]
                idx_max = ii - n_interp_rad

    #coh_max = interp_points[n_interp_rad,:,:]
    swe_est = swes[idx_max]

    #quadratic components (hardcoded for n_interp = 3)
    a = interp_points[1]
    b = -0.5*interp_points[0] + 0.5*interp_points[2]
    c = 0.5*interp_points[0] -1.0*interp_points[1] + 0.5*interp_points[2]
    dx = -b/2/c
    #handle cases where center point is not the highest
    dx = np.clip(dx, -1, 1)
    swe_est_refined = swe_est + dx*swe_del
    coh_max_refined = a + b*dx + c*dx**2
    print('swe_est_refined, coh_max_refined:', swe_est_refined, coh_max_refined)

    phi_masked = phi[~exclude_mask]
    phi_demean = np.angle(phi_masked*np.conj(np.mean(phi_masked)))
    phi_demean -= np.mean(phi_demean)

    xi_masked = xi[~exclude_mask]
    xi_demean = xi_masked-np.mean(xi_masked)

    swe_est_corr = np.mean(phi_demean*xi_demean)/np.mean(xi_demean**2)
    print('swe_est_corr:', swe_est_corr)

    def lad_cost(x, v, w):
        cost = np.sum(np.abs(x[0]*v - w))
        print('swe, cost', x[0], cost)
        return cost

    def fit_lad_swe(swe_init, xi_demean, phi_demean):
        scale = 1
        x_init = (swe_init/scale)
        bounds = ((-1./scale, 1./scale),)
        gtol = 1e-13; ftol = 1e-6;
        maxiter = 200
        minimize_options = {'gtol': gtol, 'ftol': ftol, 'maxiter': maxiter}
        opt_result = opt.minimize(lad_cost, x_init, args=(xi_demean, phi_demean), options=minimize_options, bounds=bounds)
        swe = opt_result.x[0]*scale

        print('*** Final solution ***')
        print('swe = ', swe)
        return swe
    swe_est_lad = fit_lad_swe(swe_est_corr, xi_demean, phi_demean)
    print('swe_est_lad:', swe_est_lad)


    fig = plt.figure(figsize=(fig_width_full, 4.5))
    gs = gridspec.GridSpec(ncols=6, nrows=2, figure=fig)
    fig.add_subplot(gs[1,0:3])
    plt.hist2d(xi_demean/1000., phi_demean, bins=100, cmap='Greys')

    xpts = np.asarray([np.min(xi_demean), np.max(xi_demean)])
    plt.plot(xpts/1000., xpts*swe_est_refined, color='tab:red', label=r'$\Delta\hat S_w$ = {:02.0f} mm'.format(swe_est_refined*1000.))
    plt.xlabel(r'$\~\xi$ [$rad \ mm^{-1}$]')
    plt.ylabel(r'$\~\Phi$ [rad]')
    plt.gca().yaxis.set_major_locator(plt.MultipleLocator(np.pi))
    plt.ylim((-np.pi, np.pi))
    plt.legend(loc='lower right')
    plt.title(r'2D-histogram of $\~\Phi$ vs. $\~\xi$')

    def format_func(value, tick_number):
        # find number of multiples of pi
        N = int(np.round(value / np.pi))
        if N == 0:
            return "0"
        elif N == 1:
            return r"$\pi$"
        elif N == -1:
            return r"-$\pi$"
        else:
            return r"${0}\pi$".format(N)

    plt.gca().yaxis.set_major_formatter(plt.FuncFormatter(format_func))

    fig.add_subplot(gs[1, 3:])
    plt.title('Periodogram')
    plt.plot(swes*1000., coh, 'k')
    plt.plot([swe_est_refined*1000., swe_est_refined*1000.], [0., coh_max_refined*1.1], '--', color='tab:red')
    plt.text((swe_est_refined+0.005)*1000., 0.0, r'$\Delta\hat S_w$ = {:02.0f} mm'.format(swe_est_refined*1000.) , horizontalalignment='left', verticalalignment='bottom', size=8)

    plt.xlim((-0.3*1000, 0.3*1000))
    plt.ylim((0., coh_max_refined*1.1))
    plt.xlabel(r'$\Delta$SWE [mm]')
    plt.ylabel('Spectral magnitude')

    extent = (0, 1, 0, 1)
    fig.add_subplot(gs[0, 0:2])
    vmin = -0.04; vmax=0.06
    plt.title(r'$\~\xi$')
    im = plt.imshow((xi[::-1, :].T - np.mean(xi)) /1000., vmin=vmin, vmax=vmax, extent=extent)
    plt.xlabel('Ground range position [km]')
    plt.ylabel('Azimuth position [km]')
    plt.xticks([0, 1])
    plt.yticks([0, 1])

    divider = make_axes_locatable(plt.gca())
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax, ticks = [vmin, vmax])

    fig.add_subplot(gs[0, 2:4])
    plt.title(r'$\~\Phi$')
    im = plt.imshow(np.angle(phi)[::-1, :].T, vmin=-np.pi, vmax=np.pi, cmap=cmap_gamma(), extent=extent)
    plt.xlabel('Ground range position [km]')
    plt.ylabel('Azimuth position [km]')
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    divider = make_axes_locatable(plt.gca())
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax, ticks = [-np.pi, np.pi])
    cbar.ax.set_yticklabels([r'-$\pi$', r'$\pi$'])

    fig.add_subplot(gs[0, 4:])
    plt.title(r'$\~\Phi - \Delta\hat S_w \~\xi$')
    im = plt.imshow(np.angle(phi*np.exp(-1j*swe_est_refined*(xi-np.mean(xi))))[::-1, :].T, vmin=-np.pi, vmax=np.pi, cmap=cmap_gamma(), extent=extent)
    plt.xlabel('Ground range position [km]')
    plt.ylabel('Azimuth position [km]')
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    divider = make_axes_locatable(plt.gca())
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax, ticks = [-np.pi, np.pi])
    cbar.ax.set_yticklabels([r'-$\pi$', r'$\pi$'])

    correlation= np.sum(xi_demean*phi_demean)/np.sqrt(np.sum(xi_demean**2)*np.sum(phi_demean**2))
    print('Correlation coef.:', correlation)

    plt.tight_layout()
    if plots_for_paper:
        plt.savefig(fig_dir + 'inset_example.png')

    plt.show()

    return

def monte_carlo_mpinstance(coh, xi,  swe_range=(-0.1, 0.1), swe_del = 0.002, win=(64,64), exclude_mask=None):
    np.random.seed()
    swe_render_lim = np.max(np.abs(swe_range))
    dim = coh.shape

    x1 = np.random.randn(dim[0], dim[1]).astype(np.float32) + 1j*np.random.randn(dim[0], dim[1]).astype(np.float32)
    x2 = np.random.randn(dim[0], dim[1]).astype(np.float32) + 1j*np.random.randn(dim[0], dim[1]).astype(np.float32)
    x2 = (coh*x1 + (1 - coh**2)**0.5*x2)
    phi_sim = np.conj(x1)*x2
    x1=None
    x2=None

    swe_est_refined, _ = est_swe(phi_sim, xi, swe_range=swe_range, swe_del=swe_del, win=win, exclude_mask=exclude_mask)
    return swe_est_refined

def monte_carlo(cfg, pair_name=None, plot = False):
    #monte carlo estimation of SWE variance and bias
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    win=cfg['win']
    swe_est_dir = cfg['swe_est_dir']
    swe_range = cfg['swe_range']
    swe_del = cfg['swe_del']

    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del

    if pair_name is None:
        #form list using glob and then recurse back in same function for each pair found
        swe_fnames = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe')
        pair_names = [os.path.basename(swe_fname)[0:17] for swe_fname in swe_fnames]
        print('Found', len(pair_names),'. Performing Monte Carlo on each one...')
        for pair_name in pair_names:
            print(pair_name, '...')
            monte_carlo(cfg, pair_name=pair_name, plot=plot)
        return

    swe_render_lim = 60. #mm

    max_gb = 64.
    n_iter = 40
    crop = None

    swe_fname = swe_est_dir+pair_name+'.diff.swe'
    sim_sd_fname = swe_fname + '.sim_sd'

    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')

    xi = get_xi(cfg)

    exclude_mask = get_water_mask(cfg)  # just exclude water areas since the blunders were masked out and filled already

    if crop is not None:
        exclude_mask = exclude_mask[crop[0]:crop[2], crop[1]:crop[3]]
    # get standard deviation of xi
    xi_masked = np.copy(xi)
    xi_masked[exclude_mask] = 0
    good_mask = np.ones_like(xi)
    good_mask[exclude_mask] = 0
    good_mask_filtered = uniform_filter(good_mask, win)
    xi_sd = np.sqrt(uniform_filter(xi_masked ** 2, win) / good_mask_filtered - (
                uniform_filter(xi_masked, win) / good_mask_filtered) ** 2)

    if crop is not None:
        dim = crop[2:]
    else:
        dim = par.dim
    if not os.path.exists(sim_sd_fname):
        coh = readBin(swe_fname+'.coh', par.dim, 'float32', crop=crop)

        multiprocess=True
        job_gb = 15
        nodes_max = int(max_gb/job_gb)
        #TODO fix the mp code so I can use nodes max!!!
        p = mp.ProcessingPool(multiprocess=multiprocess)
        res = p.map(lambda ii: monte_carlo_mpinstance(coh, xi,  swe_range=swe_range, swe_del=swe_del, win=win, exclude_mask=exclude_mask), np.arange(n_iter))

        swe_sims = np.asarray(res)
        swe_sd = np.var(swe_sims, axis=0)**0.5
        swe_bias = np.mean(swe_sims, axis=0)

        writeBin(sim_sd_fname, swe_sd)
        render_png(swe_sd.T * 1000, sim_sd_fname + '.png', vmin=0, vmax=10)
    else:
        swe_sd = readBin(sim_sd_fname, dim, 'float32')

    if plot:
        swe_est = readBin(swe_fname, dim, 'float32')
        coh_max = readBin(swe_fname+'.coh', dim, 'float32')
        nan_mask = (~np.isfinite(swe_est)) | (~np.isfinite(swe_sd))
        bad_mask = nan_mask | (swe_est <= swes[4]) | (swe_est >= swes[-4])
        extent = np.asarray([0, par['range_pixel_spacing']*par['range_samples']/np.sin(np.deg2rad(par['incidence_angle'])),0, par['azimuth_pixel_spacing']*par['azimuth_lines']])/1000

        plt.figure(figsize=(fig_width_full, 4.))

        plt.subplot(2,2,1)
        cmap = cm.get_cmap('RdBu')
        cmap.set_bad(color='grey')
        swe_est[bad_mask] = np.nan
        im=plt.imshow(swe_est[::-1, :].T*1000, cmap='RdBu', vmin=-swe_render_lim,vmax=swe_render_lim, extent=extent)
        plt.title(r'Estimated $\Delta$SWE')
        plt.xticks([])
        plt.yticks([])
        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, ticks = [-swe_render_lim, swe_render_lim])
        cbar.ax.get_yaxis().labelpad = 0

        plt.subplot(2,2,2)
        cmap = cm.get_cmap('viridis')
        cmap.set_bad(color='grey')
        coh_max[bad_mask]=np.nan
        im = plt.imshow(coh_max[::-1, :].T, vmin=0, vmax=1, extent=extent, cmap=cmap)
        plt.title('Residual Phase Coherence')
        plt.xticks([])

        plt.yticks([])
        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, ticks = [0, 1])
        cbar.ax.get_yaxis().labelpad = 0

        plt.subplot(2,2,3)
        xi_sd[exclude_mask]=np.nan
        im=plt.imshow(xi_sd[::-1, :].T, vmin=0, vmax=20, extent=extent, cmap=cmap)
        plt.title(r'$\xi$ std. dev.')

        plt.xticks([])
        plt.yticks([])
        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, ticks = [0, 20])
        cbar.ax.get_yaxis().labelpad = 0

        plt.subplot(2,2,4)
        swe_sd[bad_mask]=np.nan
        im=plt.imshow(swe_sd[::-1, :].T * 1000., vmin=0, vmax=10, extent=extent, cmap=cmap)
        plt.title(r'Estimated $\Delta$SWE std. dev.')

        plt.xticks([])
        plt.yticks([])
        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, ticks = [0, 10])
        cbar.ax.get_yaxis().labelpad = 0

        plt.tight_layout()
        plt.savefig(fig_dir + 'monte_carlo.png')


def est_swe_all(stack=stack_SLA27D, diff_dir = None, out_dir = None, looks_str = '3_12', swe_range=(-0.05, 0.08), swe_render_lim=None, swe_del=0.002, win=(128,128), crop=None, suffix=''):
    #START config
    multiprocess=True

    if swe_render_lim is None:
        swe_render_lim = np.max(np.abs(swe_range))

    if diff_dir is None:
        diff_dir = 'diff_'+looks_str+'/'

    if out_dir is None:
        out_dir = 'swe_est_'+looks_str+'_'+suffix+'/'
    out_dir = stack._dir + out_dir
    #END config

    ensureDir(out_dir)
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    if crop is not None:
        dim = crop[2:]
    else:
        dim = par.dim
    xi = get_xi(cfg)

    exclude_mask = get_water_mask(cfg)#just exclude water areas since the blunders were masked out and filled already
    if crop is not None:
        exclude_mask = exclude_mask[crop[0]:crop[0]+crop[2],crop[1]:crop[1]+crop[3]]

    ifg_fnames = glob.glob(stack._dir + diff_dir + '*.diff')
    ifg_fnames.sort()
    ifg_fnames = ifg_fnames

    n_ifg = len(ifg_fnames)

    #generate all the swe estimates
    p = mp.ProcessingPool(multiprocess=multiprocess)
    p.map(lambda ifg_fname: est_swe_from_file(ifg_fname, par, crop, xi,  out_dir, swe_range=swe_range, swe_render_lim = swe_render_lim, swe_del=swe_del, win=win, exclude_mask=exclude_mask), ifg_fnames)
    #render results (no multiprocess allowed)
    p = mp.ProcessingPool(multiprocess=False)
    p.map(lambda ifg_fname: est_swe_from_file_render(ifg_fname, par, crop, xi,  out_dir, swe_range=swe_range, swe_render_lim = swe_render_lim, swe_del=swe_del, win=win, exclude_mask=exclude_mask), ifg_fnames)


def mk_water_mask_SLA27D():
    rmli_dir = './inuvik/RS2_SLA27_D/full_scene/rmli_3_12/'
    date_list_init = [os.path.splitext(os.path.basename(fname))[0] for fname in glob.glob(rmli_dir+'????????.rmli')]
    date_list = []
    for date in date_list_init:
        if (int(date[4:6]) >= 6) and (int(date[4:6]) < 10):
            date_list += [date]
    mk_water_mask(rmli_dir,rmli_dir,date_list, iterations=1, SLA=True)

def mk_diff_sim(cfg, sim_run):
    stack = cfg['stack']
    looks_str = cfg['look_str']

    #END config

    sim_dir = stack._dir+'snowmodel_'+looks_str+'/'+sim_run+'/'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    xi = get_xi(cfg)
    out_dir = sim_dir
    ensureDir(out_dir)

    swe_fnames = sorted(glob.glob(sim_dir + '????????.swed.rdc'))
    for ii in np.arange(len(swe_fnames)-1):
        print('Simulating', ii+1, 'of', len(swe_fnames)-1)
        swe_m = readBin(swe_fnames[ii], par.dim, 'float32')
        swe_s = readBin(swe_fnames[ii+1], par.dim, 'float32')

        print(np.min(swe_m), np.max(swe_m))

        pair_name = (os.path.basename(swe_fnames[ii])).split('.')[0] + '_' + (os.path.basename(swe_fnames[ii+1])).split('.')[0]
        swe_diff = swe_s - swe_m

        phi = np.exp(1j*xi*swe_diff)
        writeBin(out_dir + pair_name+'.diff', phi)
        render_png(np.angle(phi.T), out_dir + pair_name+'.diff.png', cmap=cmap_gamma(), vmin=-np.pi, vmax=np.pi)
        render_png(swe_diff.T, out_dir + pair_name + '.swe.png', cmap='RdBu', vmin=-0.03, vmax=0.03)


def sim_multiplot(cfg, index_list, sim_run, plot_singles = False, plot_history = False, cache_filename = None):
    #loop through grid of window locations
    #for each window...
    #extract xi, s chips
    #decompose
    #compute estimate directly and by component and compare
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    win=cfg['win']
    step = win//4
    #END config

    sim_dir = stack._dir+'snowmodel_'+looks_str+'/'+sim_run+'/'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    xi_all = get_xi(cfg)

    #define mask to exclude from estimation
    exclude_mask_all = get_water_mask(cfg)

    grid_sz = par.dim // step
    exclude_mask_grid = np.zeros(grid_sz, dtype=np.bool)
    for ii in np.arange(grid_sz[0]):
        #print(ii+1, 'of', grid_sz[0])
        for jj in np.arange(grid_sz[1]):
            exclude_mask = exclude_mask_all[ii * step[0]:ii * step[0] + win[0], jj * step[1]:jj * step[1] + win[1]]
            exclude_mask_grid[ii, jj] = exclude_mask[exclude_mask.shape[0]//2, exclude_mask.shape[1]//2]

    swe_fnames = sorted(glob.glob(sim_dir + '????????.swed.rdc'))

    #trim the index list to leave out summer igrams (just take Oct 1 to Mar 31)
    #Oct 1 DOY = 274, Mar 31 DOY = 90
    new_idx_list = []
    dt_list = []
    for idx in index_list:
        root_name = (os.path.basename(swe_fnames[idx])).split('.')[0]
        doy_m = datetime.strptime(root_name, '%Y%m%d').timetuple().tm_yday
        if (doy_m >=274) or ((doy_m + 24) <= 90):
            #winter season igram so keep index
            new_idx_list.append(idx)
            dt_list.append(datetime.strptime(root_name, '%Y%m%d'))
    index_list = new_idx_list

    grid_sz = par.dim // step
    num_igrams = len(index_list)
    err_all, swe_mean_all =sim_swe_error_all(sim_run, swe_fnames, index_list, xi_all, exclude_mask_all, par, win, step, plot = plot_singles)
    err_all[:,exclude_mask_grid] = np.nan
    swe_mean_all[:,exclude_mask_grid] = np.nan

    #remove outliers
    index_list_outliers = [7]

    outliers_mask = np.asarray([idx in index_list_outliers for idx in index_list])
    err_nooutliers = err_all[~outliers_mask,:,:]
    swe_mean_nooutliers = swe_mean_all[~outliers_mask, :, :]
    dt_list = [dt_list[ii] for ii in np.arange(len(dt_list)) if not outliers_mask[ii]]

    err_mean = np.mean(err_nooutliers, axis=0)
    err_rms = np.mean(err_nooutliers**2, axis=0)**0.5
    swe_mean = np.mean(swe_mean_all, axis=0)

    err_spatial_mean = np.nanmean(err_nooutliers, axis=(1,2))
    err_spatial_rms = np.nanmean(err_nooutliers**2, axis=(1, 2))**0.5

    swe_spatial_mean = np.nanmean(swe_mean_nooutliers, axis=(1,2))
    extent = np.asarray(
        [0, par['range_pixel_spacing'] * par['range_samples'] / np.sin(np.deg2rad(par['incidence_angle'])), 0,
         par['azimuth_pixel_spacing'] * par['azimuth_lines']]) / 1000

    if plot_history:
        fig = plt.figure(figsize=(fig_width_full, 1.8))
        # #lim = np.max(np.abs((np.percentile(swe_mean[np.isfinite(swe_mean)], 5), np.percentile(swe_mean[np.isfinite(swe_mean)], 95)))) * 1000
        #
        lim = 12
        cmap = cm.get_cmap('RdBu')
        cmap.set_bad(color='grey')
        gs = gridspec.GridSpec(ncols=3, nrows=1, figure=fig)

        fig.add_subplot(gs[0,0])

        if sim_run == 'bare':
            lim = 2.
        elif sim_run == 'nalcms':
            lim = 100.

        im = plt.imshow(err_mean.T[:,::-1]*1000, vmin = -lim, vmax=lim, cmap = cmap,extent=extent)
        plt.title(r'Temporal mean $\Delta$SWE error')
        plt.xlabel('Ground range position [km]')
        plt.ylabel('Azimuth\nposition [km]')

        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, ticks = [-lim, lim])
        cbar.ax.get_yaxis().labelpad = 10
        cbar.ax.set_ylabel('mm SWE', rotation=90)

        fig.add_subplot(gs[0,1])
        if sim_run == 'bare':
            range = (-5, 5)
        elif sim_run == 'nalcms':
            range = (-100, 100)

        plt.hist(err_all[np.isfinite(err_all)].flatten() * 1000., bins=50, range=range)
        plt.title(r'Histogram of all $\Delta$SWE errors')
        plt.xlabel(r'$\Delta$SWE error [mm]')
        plt.ylabel('Frequency')
        ylim = plt.gca().get_ylim()
        plt.gca().axes.yaxis.set_ticks(ylim)
        plt.gca().set_yticklabels(['0', 'max'])
        global_mean_err = np.nanmean(err_nooutliers)
        global_rms_err = np.nanmean(err_nooutliers ** 2) ** 0.5
        ylim = plt.gca().get_ylim()
        xlim = plt.gca().get_xlim()
        plt.text((xlim[0] + 0.02*(xlim[1]-xlim[0])), ylim[0] + 0.95*(ylim[1]-ylim[0]), r'Mean $\Delta$SWE' '\nerror:' '\n{:2.1f} mm'.format(global_mean_err*1000),
                 horizontalalignment='left', verticalalignment='top', size=6)
        plt.text((xlim[0] + 0.60*(xlim[1]-xlim[0])), ylim[0] + 0.95*(ylim[1]-ylim[0]), r'RMS $\Delta$SWE' '\nerror:' '\n{:2.1f} mm'.format(global_rms_err*1000),
                 horizontalalignment='left', verticalalignment='top', size=6)

        fig.add_subplot(gs[0,2])
        plt.hist(err_spatial_mean*1000, bins=50, range=[None, 20])
        plt.title('\nHistogram of per-interval\n' r'spatial mean $\Delta$SWE error')
        plt.xlabel(r'Mean $\Delta$SWE error [mm]')
        plt.xticks([-20, -10, 0, 10, 20])
        plt.ylabel('Frequency')
        ylim = plt.gca().get_ylim()
        plt.gca().axes.yaxis.set_ticks(ylim)
        plt.gca().set_yticklabels(['0', 'max'])

        plt.tight_layout()
        plt.savefig(fig_dir + 'sim_fullhistory_' + sim_run + '.png')

        #BSI analysis looks at any correspondence between cumulative BSI and simulated BIAS
        fractional_day = 0.63
        era5_data = get_inuvik_RS2_SLA27D_era5()
        dates = [os.path.basename(fname)[0:8] for fname in swe_fnames]
        era5_data_pairwise = snow.get_era5_pairwise_data(era5_data, dates, fractional_day)
        bsi = np.diagonal(era5_data_pairwise['cum_bsi'], offset=1)[index_list]
        plt.figure()
        plt.subplot(3,1,1)
        plt.plot(dt_list, err_spatial_mean*1000)
        plt.xlabel('Date')
        plt.ylabel('Mean SWE Bias (mm)')
        plt.subplot(3,1,2)
        plt.plot(dt_list, bsi[~outliers_mask])
        plt.xlabel('Date')
        plt.xlabel('Cumulative BSI')
        plt.subplot(3, 1, 3)
        plt.plot(bsi[~outliers_mask], err_spatial_mean*1000, '*')
        plt.xlabel('Cumulative BSI')
        plt.ylabel('Mean SWE Bias (mm)')

def sim_swe_error_all(sim_run, swe_fnames, index_list, xi_all, exclude_mask_all, par, win, step, plot = False):
    num_igrams = len(index_list)
    grid_sz = par.dim // step
    err_all = np.zeros((num_igrams, grid_sz[0], grid_sz[1]))
    swe_mean_all = np.zeros((num_igrams, grid_sz[0], grid_sz[1]))
    for ii in np.arange(num_igrams):
        idx = index_list[ii]
        swe_m = readBin(swe_fnames[idx], par.dim, 'float32')
        swe_s = readBin(swe_fnames[idx + 1], par.dim, 'float32')
        pair_name = (os.path.basename(swe_fnames[idx])).split('.')[0] + '_' + (os.path.basename(swe_fnames[idx + 1])).split('.')[0]
        s_all = swe_s - swe_m
        print(pair_name, ':')
        err_all[ii,:,:], swe_mean_all[ii,:, :] = sim_swe_error_single(sim_run, s_all, xi_all, exclude_mask_all, par, win, step, plot=plot)
    return err_all, swe_mean_all

def sim_swe_error_single(sim_run, s_all, xi_all, exclude_mask_all, par, win, step, plot = False):
    #estimation grid
    grid_sz = par.dim//step
    err_arr = np.zeros(grid_sz)
    err1_arr = np.zeros(grid_sz)
    err2_arr = np.zeros(grid_sz)

    s_mean_arr = np.zeros(grid_sz)
    s_est_arr = np.zeros(grid_sz)

    exclude_mask_grid = np.zeros(grid_sz, dtype=np.bool)
    for ii in np.arange(grid_sz[0]):

        for jj in np.arange(grid_sz[1]):
            xi = xi_all[ii*step[0]:ii*step[0]+win[0], jj*step[1]:jj*step[1]+win[1]]
            s = s_all[ii*step[0]:ii*step[0]+win[0], jj*step[1]:jj*step[1]+win[1]]
            exclude_mask = exclude_mask_all[ii * step[0]:ii * step[0] + win[0], jj * step[1]:jj * step[1] + win[1]]
            xi = xi[~exclude_mask]
            s = s[~exclude_mask]

            xi_mean = np.mean(xi)
            xi_tild = xi-xi_mean

            s_mean = np.mean(s)
            s_tild = s-s_mean
            phis = xi*s
            phis_tild = phis - np.mean(phis)

            s_est = np.mean(xi_tild*phis_tild)/np.mean(xi_tild**2)

            err = s_est - s_mean

            err1 = xi_mean*np.mean(xi_tild*s_tild)/np.mean(xi_tild**2)

            err2 = np.mean(xi_tild**2*s_tild)/np.mean(xi_tild**2)
            err_arr[ii,jj] = err
            err1_arr[ii, jj] = err1
            err2_arr[ii, jj] = err2

            s_mean_arr[ii, jj] = s_mean
            s_est_arr[ii,jj] = s_est
            exclude_mask_grid[ii, jj] = exclude_mask[exclude_mask.shape[0] // 2, exclude_mask.shape[1] // 2]

            x=1

    exclude_mask_result = ~np.isfinite(s_est_arr) | exclude_mask_grid
    err_arr[exclude_mask_result] = np.nan
    mean_err = np.mean(err_arr[~exclude_mask_result])
    rms_err = np.mean(err_arr[~exclude_mask_result]**2)**0.5

    print('mean_err [mm], rms_err [mm]:', mean_err*1000., rms_err*1000.)
    if plot is True:
        #bare settings
        if sim_run == 'bare':
            lim_delSWE_sim = np.asarray([-20, 20.])
            lim_delSWE_error = np.asarray([-5, 5.])
            lim_2dhistx = np.asarray([11., 15.])
            lim_2dhisty = np.asarray([11., 15.])
        elif sim_run == 'nalcms':
            lim_delSWE_sim = np.asarray([-20, 20.])
            lim_delSWE_error = np.asarray([-100, 100.])
            lim_2dhistx = np.asarray([5, 15])
            lim_2dhisty = np.asarray([-10, 30])

        fig = plt.figure(figsize=(fig_width, 6))

        gs = gridspec.GridSpec(ncols=1, nrows=3, figure=fig)
        fig.add_subplot(gs[0,0])
        plt.title(r'Simulated $\Delta$SWE')
        s_display = np.copy(s_all)
        extent = np.asarray([0, par['range_pixel_spacing']*par['range_samples']/np.sin(np.deg2rad(par['incidence_angle'])),0, par['azimuth_pixel_spacing']*par['azimuth_lines']])/1000
        s_display[exclude_mask_all] = np.nan
        cmap = cm.get_cmap('RdBu')
        cmap.set_bad(color='grey')
        im = plt.imshow((s_display.T)[:,::-1]*1000, vmin=lim_delSWE_sim[0], vmax=lim_delSWE_sim[1], cmap=cmap, extent = extent)
        plt.xlabel('Ground range position [km]' '\n')
        plt.ylabel('Azimuth position [km]')
        plt.yticks([0,5,10])
        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, ticks=[lim_delSWE_sim[0], lim_delSWE_sim[1]])
        cbar.ax.get_yaxis().labelpad = 0
        cbar.ax.set_ylabel('mm SWE', rotation=90)

        fig.add_subplot(gs[1, 0])
        plt.title(r'$\Delta$SWE error')
        im = plt.imshow((err_arr.T)[:,::-1]*1000, vmin=lim_delSWE_error[0], vmax=lim_delSWE_error[1], cmap=cmap, extent = extent)
        plt.xlabel('Ground range position [km]' '\n')
        plt.ylabel('Azimuth position [km]')
        plt.yticks([0,5,10])
        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, ticks=[lim_delSWE_error[0], lim_delSWE_error[1]])
        cbar.ax.get_yaxis().labelpad = 0
        cbar.ax.set_ylabel('mm SWE', rotation=90)

        fig.add_subplot(gs[2, 0])
        plt.hist(err1_arr.flatten()*1000, bins=100, range = lim_delSWE_error)
        plt.title(r'Histogram of $\Delta$SWE error')
        ylim = plt.gca().get_ylim()
        plt.gca().axes.yaxis.set_ticks(ylim)
        plt.gca().set_yticklabels(['0', 'max'])
        plt.xlabel(r'$\Delta$SWE error [mm]' '\n')
        plt.ylabel('Frequency')
        ylim = plt.gca().get_ylim()
        plt.text((lim_delSWE_error[0] + 0.1), ylim[0] + 0.6*(ylim[1]-ylim[0]), r'Mean $\Delta$SWE' '\nerror = {:2.1f} mm'.format(mean_err*1000),
                 horizontalalignment='left', verticalalignment='top', size=8)
        plt.text((lim_delSWE_error[0] + 0.1), ylim[0] + 0.9*(ylim[1]-ylim[0]), r'RMS $\Delta$SWE' '\nerror = {:2.1f} mm'.format(rms_err*1000),
                 horizontalalignment='left', verticalalignment='top', size=8)

        plt.tight_layout()
        plt.savefig(fig_dir + 'sim_multiplot_' + sim_run + '.png')
    return err_arr, s_mean_arr

def rdcras_to_geotiff(inras, outtif=None, stack = stack_SLA27D, looks_str = '3_12',interp_mode=1):
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    dem_par = DEM_Par(stack._dir + 'dem_' + looks_str + '/seg.dem_par')
    gc_table = stack._dir + 'dem_' + looks_str + '/gc_fine'

    rootname = os.path.splitext(inras)[0]
    ext = os.path.splitext(inras)[1]
    if ext == '.png':
        result1 = exec(['convert', inras, rootname + '.ras'])
        inras = rootname + '.ras'
    georas = rootname + '.geo.ras'
    if outtif is None:
        outtif = rootname + '.geo.tif'
    res1 = run('geocode_back', inras, par.dim[0], gc_table, georas, dem_par.dim[0], None, interp_mode, 2)
    res2 = run('data2geotiff', dem_par, georas, 0,outtif)

def estimation_drill_down_plot(cfg):
    # make drill-down fig
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    xi = get_xi(cfg)
    exclude_mask = get_water_mask(cfg)  # just exclude water areas since the blunders were masked out and filled already
    ifg_fname = stack._dir + 'diff_' + looks_str + '/20120317_20120410.diff'
    phi = readBin(ifg_fname, par.dim, 'complex64')
    swe_est_clicky_func([1412, 946], cfg, phi, xi, exclude_mask, plots_for_paper=True)

def inuvik_climate_normals_plot():
    #data taken from: https://climate.weather.gc.ca/climate_normals/results_1981_2010_e.html?searchType=stnName&txtStationName=inuvik&searchMethod=contains&txtCentralLatMin=0&txtCentralLatSec=0&txtCentralLongMin=0&txtCentralLongSec=0&stnID=1669&dispBack=1
    months = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']
    months = np.arange(12)+1
    months_label = ['Jan', '', '', 'Apr', '', '', 'Jul', '', '', 'Oct', '', '']
    months_label = ['Jan','','Mar','','May','','Jul','','Sep','','Nov','']
    daily_maximum_temperature = [-22.8,-20.9,-16.8,-6.3,5.2,17.7,19.5,16.0,7.9,-4.3,-17.1,-20.0]
    snowfall = np.asarray([15.4,16.4,14.8,11.7,14.2,1.9,0.0,3.1,10.4,30.1,21.3,19.4])*10 #mm

    print('Mean annual snowfall [mm]: ', np.sum(snowfall))
    plt.figure(figsize=(fig_width, 2))
    plt.title('Inuvik climate normals')
    ax1 = plt.gca()
    blue = 'tab:blue'
    lns1 = plt.plot(months, snowfall, color=blue, label = 'Monthly\nsnowfall')
    ax1.tick_params(axis='y', labelcolor=blue)
    plt.ylim((0,None))
    plt.xlim((1, 12))
    plt.ylabel('Snowfall [mm]')
    ax1.set_ylabel('Snowfall [mm]', color=blue)
    plt.xticks(months, months_label)
    ax2 = plt.gca().twinx()  # instantiate a second axes that shares the same x-axis

    ax2.set_ylabel('Daily max temp [$^\circ$C]', color='k')  # we already handled the x-label with ax1
    lns2 = ax2.plot(months, daily_maximum_temperature,label = 'Daily\nmax temp', color='k')
    lns2 = ax2.plot(months, np.zeros(12), '--', color='k')
    ax2.tick_params(axis='y', labelcolor='k')

    plt.xticks(months, months_label)
    # added these three lines
    lns = lns1 + lns2
    labs = [l.get_label() for l in lns]
    plt.tight_layout()  # otherwise the right y-label is slightly clipped
    plt.savefig(fig_dir + 'inuvik_climate_normals.png')

def other_phase_sources(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    win=cfg['win']
    step = win//4
    k = 1e-5 #typical bulk delay difference as per Lin et all 2010
    beta_ds = 0.010 #reference case changed to better align with heave reference case

    del_t = 24./365.25
    topo_dir = stack._dir + 'dem_' + looks_str + '/topo/'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    lamb = constants.c/par['radar_frequency']
    xi_all = get_xi(cfg)
    exclude_mask_all = get_water_mask(cfg)  # just exclude water areas since the blunders were masked out and filled already
    dem_fn =   stack._dir + 'dem_' + looks_str + '/seg.dem.rdc'
    dem_all = readBin(dem_fn, par.dim, 'float32')

    heave_phase_sensitivity_all = readBin(topo_dir + 'dphi_dheave.rdc', par.dim, 'float32')
    downslope_phase_sensitivity_all = readBin(topo_dir + 'dphi_ddownslope_slope_scaled.rdc', par.dim, 'float32')

    grid_sz = par.dim//step
    stat_atm_corr = np.zeros(grid_sz)
    heave_corr = np.zeros(grid_sz)
    downslope_corr = np.zeros(grid_sz)
    exclude_mask_grid = np.zeros(grid_sz, dtype=bool)

    for ii in np.arange(grid_sz[0]):
        #print(ii+1, 'of', grid_sz[0])
        for jj in np.arange(grid_sz[1]):
            xi = xi_all[ii*step[0]:ii*step[0]+win[0], jj*step[1]:jj*step[1]+win[1]]
            dem = dem_all[ii * step[0]:ii * step[0] + win[0], jj * step[1]:jj * step[1] + win[1]]
            heave_phase_sensitivity = heave_phase_sensitivity_all[ii * step[0]:ii * step[0] + win[0], jj * step[1]:jj * step[1] + win[1]]
            downslope_phase_sensitivity = downslope_phase_sensitivity_all[ii * step[0]:ii * step[0] + win[0],
                                      jj * step[1]:jj * step[1] + win[1]]

            exclude_mask = exclude_mask_all[ii * step[0]:ii * step[0] + win[0], jj * step[1]:jj * step[1] + win[1]]
            xi = xi[~exclude_mask]
            xi_mean = np.mean(xi)
            xi_tild = xi-xi_mean

            dem = dem[~exclude_mask]
            heave_phase_sensitivity = heave_phase_sensitivity[~exclude_mask]
            downslope_phase_sensitivity = downslope_phase_sensitivity[~exclude_mask]
            exclude_mask_grid[ii, jj] = exclude_mask[exclude_mask.shape[0]//2, exclude_mask.shape[1]//2]
            stat_atm_corr[ii, jj] = 4*np.pi/lamb*k*np.mean(xi_tild*dem)/np.mean(xi_tild**2)
            heave_corr[ii, jj] = 0.01*np.mean(xi_tild * heave_phase_sensitivity) / np.mean(xi_tild ** 2)
            downslope_corr[ii, jj] = del_t*beta_ds*np.mean(xi_tild * downslope_phase_sensitivity) / np.mean(xi_tild ** 2)

    stat_atm_corr[exclude_mask_grid] = np.nan
    heave_corr[exclude_mask_grid] = np.nan
    downslope_corr[exclude_mask_grid] = np.nan

    extent = np.asarray([0, par['range_pixel_spacing']*par['range_samples']/np.sin(np.deg2rad(par['incidence_angle'])),0, par['azimuth_pixel_spacing']*par['azimuth_lines']])/1000

    fig = plt.figure(figsize=(fig_width_full, 6))
    gs = gridspec.GridSpec(nrows=30, ncols=30, figure=fig)
    fig.add_subplot(gs[0:8, 0:16])
    lim = 3.
    cmap = cm.get_cmap('RdBu')
    cmap.set_bad(color='grey')
    im = plt.imshow(stat_atm_corr.T[:,::-1]*1000, vmin = -lim, vmax = lim, cmap = cmap, extent=extent)
    plt.title(r'$\Delta$SWE error caused by $\Delta K = 10^{-5}$')
    plt.xticks([])
    plt.yticks([])
    divider = make_axes_locatable(plt.gca())
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax, ticks=[-lim, lim])
    cbar.ax.get_yaxis().labelpad = 0

    fig.add_subplot(gs[1:6, 20:30])
    plt.hist(stat_atm_corr[np.isfinite(stat_atm_corr)].flatten()*1000.,bins=50, range = [-3., 3.])
    plt.title('\n' r'Histogram of $\Delta$SWE error' '\n' r'caused by $\Delta K = 10^{-5}$')
    plt.xlim([-3,3])
    ylim = plt.gca().get_ylim()
    plt.gca().axes.yaxis.set_ticks(ylim)
    plt.gca().set_yticklabels(['0', 'max'])
    plt.xlabel(r'$\Delta$SWE error [mm]')

    ylim = plt.gca().get_ylim()
    xlim = plt.gca().get_xlim()
    mean_err_sa = np.nanmean(stat_atm_corr)
    rms_err_sa = np.nanmean(stat_atm_corr**2)**0.5
    plt.text(xlim[0] + 0.02 * (xlim[1] - xlim[0]), ylim[0] + 0.95 * (ylim[1] - ylim[0]),
             r'Mean $\Delta$SWE error:' '\n{:2.2f} mm'.format(mean_err_sa * 1000),
             horizontalalignment='left', verticalalignment='top', size=8)
    plt.text(xlim[0] + 0.02 * (xlim[1] - xlim[0]), ylim[0] + 0.5 * (ylim[1] - ylim[0]),
             r'RMS $\Delta$SWE error:' '\n{:2.1f} mm'.format(rms_err_sa * 1000),
             horizontalalignment='left', verticalalignment='top', size=8)

    fig.add_subplot(gs[10:18, 0:16])

    lim = 10. #mm
    im = plt.imshow(heave_corr.T[:,::-1]*1000, vmin = -lim, vmax = lim, cmap = cmap, extent=extent)
    plt.title(r'$\Delta$SWE error per 10 mm surface heave')

    plt.xticks([])

    plt.yticks([])
    divider = make_axes_locatable(plt.gca())
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax, ticks=[-lim, lim])
    cbar.ax.get_yaxis().labelpad = 0

    fig.add_subplot(gs[11:16, 20:30])
    plt.hist(heave_corr[np.isfinite(heave_corr)].flatten()*1000.,bins=50, range = [7.5, 10.])
    plt.title('\n' r'Histogram of $\Delta$SWE error' '\n' r'per 10 mm surface heave')
    plt.xlim([7.5, 10.])
    ylim = plt.gca().get_ylim()
    plt.gca().axes.yaxis.set_ticks(ylim)
    plt.gca().set_yticklabels(['0', 'max'])
    plt.xlabel(r'$\Delta$SWE error [mm]')

    ylim = plt.gca().get_ylim()
    xlim = plt.gca().get_xlim()
    mean_err_heave = np.nanmean(heave_corr)
    rms_err_heave = np.nanmean(heave_corr**2)**0.5
    plt.text(xlim[0] + 0.02 * (xlim[1] - xlim[0]), ylim[0] + 0.95 * (ylim[1] - ylim[0]),
             r'Mean $\Delta$SWE error:' '\n{:2.1f} mm'.format(mean_err_heave*1000),
             horizontalalignment='left', verticalalignment='top', size=8)
    plt.text(xlim[0] + 0.02 * (xlim[1] - xlim[0]), ylim[0] + 0.5 * (ylim[1] - ylim[0]),
             r'RMS $\Delta$SWE error:' '\n{:2.1f} mm'.format(rms_err_heave*1000),
             horizontalalignment='left', verticalalignment='top', size=8)

    fig.add_subplot(gs[20:28, 0:16])
    lim = 2.5 #mm
    lim = 0.7
    im = plt.imshow(downslope_corr.T[:,::-1]*1000, vmin = -lim, vmax = lim, cmap = cmap, extent=extent)
    plt.title(r'$\Delta$SWE error for reference solifluction $case$')
    plt.xticks([])
    plt.yticks([])
    divider = make_axes_locatable(plt.gca())
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax, ticks=[-lim, lim])
    cbar.ax.get_yaxis().labelpad = 0

    fig.add_subplot(gs[21:26, 20:30])
    plt.hist(downslope_corr[np.isfinite(downslope_corr)].flatten()*1000,bins=50, range=[1/4.,3/4.])
    plt.title('\n' r'Histogram of $\Delta$SWE error' '\n' r'for reference solifluction $case$')
    plt.xlim([1/4.,3/4.])
    ylim = plt.gca().get_ylim()
    plt.gca().axes.yaxis.set_ticks(ylim)
    plt.gca().set_yticklabels(['0', 'max'])
    plt.xlabel(r'$\Delta$SWE error [mm]')

    ylim = plt.gca().get_ylim()
    xlim = plt.gca().get_xlim()
    mean_err_downslope = np.nanmean(downslope_corr)
    rms_err_downslope = np.nanmean(downslope_corr**2)**0.5
    plt.text(xlim[0] + 0.02 * (xlim[1] - xlim[0]), ylim[0] + 0.95 * (ylim[1] - ylim[0]),
             r'Mean $\Delta$SWE error:' '\n{:2.1f} mm'.format(mean_err_downslope*1000.),
             horizontalalignment='left', verticalalignment='top', size=8)
    plt.text(xlim[0] + 0.02 * (xlim[1] - xlim[0]), ylim[0] + 0.5 * (ylim[1] - ylim[0]),
             r'RMS $\Delta$SWE error:' '\n{:2.1f} mm'.format(rms_err_downslope*1000.),
             horizontalalignment='left', verticalalignment='top', size=8)
    plt.tight_layout()
    plt.savefig(fig_dir + 'other_phase_sources.png')


def plot_all_correlations_with_sim(cfg):
    win = cfg['win']
    swe_est_dir = cfg['swe_est_dir']
    #corr_bare_bare = correlation_with_sim(cfg['stack']._dir+'swe_est_3_12_sim_bare_win500/', 'bare', win)
    corr_real_bare = correlation_with_sim(swe_est_dir, 'bare', win)
    #corr_nalcms_nalcms = correlation_with_sim(cfg['stack']._dir+'swe_est_3_12_sim_nalcms_win500/', 'nalcms', win)
    corr_real_nalcms = correlation_with_sim(swe_est_dir, 'nalcms', win)

    fig = plt.figure(figsize=(fig_width_full, 4))
    nrow = 2; ncol=2
    plt.subplot(nrow,ncol,1)
    #plt.hist(corr_bare_bare, bins=15, range=(-1,1))
    plt.title('Estimated (from simulated phase) vs.\n' r'SnowModel $\Delta$SWE (''Bare'')')
    plt.xlabel('Correlation Coef.\n(a)')
    plt.ylabel('Frequency')
    plt.subplot(nrow,ncol,2)
    #plt.hist(corr_nalcms_nalcms, bins=15, range=(-1,1))
    plt.title('Estimated (from simulated phase) vs.\n' r'SnowModel $\Delta$SWE (''NALCMS 2015'')')
    plt.xlabel('Correlation Coef.\n(b)')
    plt.ylabel('Frequency')
    plt.subplot(nrow,ncol,3)
    plt.hist(corr_real_bare, bins=15, range=(-1,1))
    plt.title('Estimated (from RS2_SLA) vs.\n''SnowModel $\Delta$SWE (''Bare'')')
    plt.xlabel('Correlation Coef.\n(c)')
    plt.ylabel('Frequency')
    plt.subplot(nrow,ncol,4)
    plt.hist(corr_real_nalcms, bins=15, range=(-1,1))
    plt.title('Estimated (from RS2_SLA) vs.\n''SnowModel $\Delta$SWE (''Bare'')')
    plt.xlabel('Correlation Coef.\n(d)')
    plt.ylabel('Frequency')
    plt.tight_layout()
    plt.savefig(fig_dir+'all_correlation_with_sim.png')

def correlation_with_sim(swe_est_dir, sim_run, win):
    #plot all results on a yearly basis starting with July 1
    stack = stack_SLA27D
    looks_str = '3_12'
    sim_dir = stack._dir + 'snowmodel_' + looks_str + '/' + sim_run + '/'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    start_dates = ['20111001', '20121001', '20131001', '20141001', '20151001', '20161001', '20171001', '20181001',
                   '20191001', '20201001']
    end_dates = ['20120331', '20130331', '20140331', '20150331', '20160331', '20170331', '20180331', '20190331',
                   '20200331', '20210331']
    swe_range = (-0.05, 0.08)
    swe_del = 0.002
    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del
    pair_names = []
    ddays = []
    #get list of conforming pairs and their ddays (usually 24 but sometimes 48)
    for ii in np.arange(len(start_dates)):
        start_dt = datetime.strptime(start_dates[ii], '%Y%m%d')
        end_dt = datetime.strptime(end_dates[ii], '%Y%m%d')
        swe_fnames_all = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe')
        pair_names_all = [os.path.basename(swe_fname)[0:17] for swe_fname in swe_fnames_all]
        pair_names_all.sort()

        for pair_name in pair_names_all:
            master_dt = datetime.strptime((pair_name[0:8]).split('.')[0], '%Y%m%d')
            slave_dt = datetime.strptime((pair_name[9:17]).split('.')[0], '%Y%m%d')
            if (master_dt > start_dt) and (slave_dt < end_dt) and exists(sim_dir+pair_name[0:8]+'.swed.rdc') and exists(sim_dir+pair_name[9:17]+'.swed.rdc'):
                pair_names.append(pair_name)
                ddays.append(int(np.round((slave_dt - master_dt).total_seconds() / 86400)))

    n_pairs = len(pair_names)
    corrs = np.zeros(n_pairs)
    for kk in np.arange(n_pairs):
        print(kk + 1, 'of', n_pairs, '...')
        # get the sim swe
        sim_swe_m = readBin(sim_dir + pair_names[kk][0:8] + '.swed.rdc', par.dim, 'float32')
        sim_swe_s = readBin(sim_dir + pair_names[kk][9:] + '.swed.rdc', par.dim, 'float32')
        sim_swe_diff = sim_swe_s - sim_swe_m
        sim_swe_diff = uniform_filter(sim_swe_diff, win)

        # get the estimated swe
        swe_est = readBin(swe_est_dir + pair_names[kk] + '.diff.swe', par.dim, 'float32')
        nan_mask = (~np.isfinite(swe_est))
        bad_mask = nan_mask | (swe_est <= swes[4]) | (swe_est >= swes[-4])
        swe_est[bad_mask] = np.nan

        mask = np.isfinite(swe_est)
        corrs[kk] = np.corrcoef(swe_est[mask].flatten(), sim_swe_diff[mask].flatten())[0, 1]

    return corrs

def end_of_season_total_swe_maps(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    swe_est_dir = cfg['swe_est_dir']

    swe_render_lim = 300.  # mm

    end_of_season_dates = ['20130312', '20150302', '20160320', '20170315', '20180310', '20190329', '20200323']
    cmap_rdbu = cm.get_cmap('RdBu')
    cmap_rdbu.set_bad(color='grey')
    extent = np.asarray(
        [0, par['range_pixel_spacing'] * par['range_samples'] / np.sin(np.deg2rad(par['incidence_angle'])), 0,
         par['azimuth_pixel_spacing'] * par['azimuth_lines']]) / 1000
    plt.figure(figsize=(fig_width, 4))
    nrow = 4; ncol=2
    for ii in np.arange(len(end_of_season_dates)):
        plt.subplot(nrow, ncol, ii+1)
        plt.title(end_of_season_dates[ii])
        swe_fname = swe_est_dir + end_of_season_dates[ii]+'.swe'
        swe = readBin(swe_fname, par.dim, 'float32')
        if ii == 0:
            swe_mean = np.copy(swe)
        else:
            swe_mean += swe
        plt.imshow(swe.T[:,::-1]*1000, vmin=-swe_render_lim, vmax=swe_render_lim, cmap=cmap_rdbu, extent=extent)
    plt.tight_layout()
    plt.savefig(fig_dir+'total_swe_maps.png')
    swe_mean/=len(end_of_season_dates)
    plt.figure()
    plt.imshow(swe_mean.T[:, ::-1] * 1000, vmin=-200, vmax=200, cmap=cmap_rdbu, extent=extent)

def plot_validation_scatter_point(cfg, pos, plot_date, insitu_swe, insitu_swesd=None, label=None, marker=None):
    #pos is [lon, lat]
    print(label, ':', pos)
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    win = cfg['win']
    swe_est_dir = cfg['swe_est_dir']
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')

    xi_all = get_xi(cfg)
    exclude_mask = get_water_mask(cfg) #just exclude water areas since the blunders were masked out and filled already
    xi_all[exclude_mask]=np.nan

    rg_idx, az_idx = get_sarpix(pos[0], pos[1])
    swe_fname_list = glob.glob(swe_est_dir+'????????.swe')
    swe_fname_list.sort()

    #find closest in time swe maps that bracket
    dates_list = [os.path.basename(fname)[0:8] for fname in swe_fname_list]
    dt_list = [datetime.strptime(date, '%Y%m%d') for date in dates_list]
    plot_dt = datetime.strptime(plot_date, '%Y%m%d')
    dt_diff = np.asarray([(dt_list_-plot_dt).total_seconds() for dt_list_ in dt_list])
    idx_nearest = np.argmin(np.abs(dt_diff))
    if dt_list[idx_nearest] < plot_dt:
        idx_pre = idx_nearest
    else:
        idx_pre = idx_nearest-1
    idx_post = idx_pre+1
    swe_pre_all = readBin(swe_fname_list[idx_pre], par.dim, 'float32')
    swe_post_all = readBin(swe_fname_list[idx_post], par.dim, 'float32')
    swe_pre = swe_pre_all[rg_idx, az_idx]
    swe_post = swe_post_all[rg_idx, az_idx]

    xi_win = xi_all[rg_idx-win[0]//2: rg_idx+win[0]//2, az_idx-win[1]//2: az_idx+win[1]//2]
    xi_sd = np.nanstd(xi_win)
    good_count = np.sum(np.isfinite(xi_win))

    dt_fraction = (plot_dt - dt_list[idx_pre]).total_seconds()/(dt_list[idx_post] - dt_list[idx_pre]).total_seconds()
    print(good_count, win[0]*win[1], xi_sd)
    swe_interp = (1-dt_fraction)*swe_pre + dt_fraction*swe_post

    swesd_pre_all = readBin(swe_fname_list[idx_pre]+'_sd', par.dim, 'float32')
    swesd_post_all = readBin(swe_fname_list[idx_post]+'_sd', par.dim, 'float32')
    swesd_pre = swesd_pre_all[rg_idx, az_idx]
    swesd_post = swesd_post_all[rg_idx, az_idx]
    swesd_interp = ((1-dt_fraction)*swesd_pre**2 + dt_fraction*swesd_post**2)**0.5
    yerr = 2*swesd_interp * 1000

    xerr=None
    if insitu_swesd is not None:
        xerr = 2*insitu_swesd*1000
    plt.errorbar([insitu_swe * 1000], [swe_interp*1000], xerr=xerr, yerr=yerr, capsize=2, label=label, marker=marker, linestyle="None")

def swe_maps_all(cfg):
    #plot all results on a yearly basis starting with July 1
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    swe_est_dir = cfg['swe_est_dir']
    swe_range = cfg['swe_range']
    swe_del = cfg['swe_del']
    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')

    cummulative = False
    exclude_mask = get_water_mask(cfg)

    if cummulative:
        start_dates = ['20121001', '20131001', '20141001', '20151001', '20161001', '20171001', '20181001',
                       '20191001']
        end_dates = ['20130331', '20140331', '20150331', '20160331', '20170331', '20180331', '20190331',
                       '20200331']
    else:
        start_dates = ['20111001', '20121001', '20131001', '20141001', '20151001', '20161001', '20171001', '20181001',
                       '20191001', '20201001']
        end_dates = ['20120331', '20130331', '20140331', '20150331', '20160331', '20170331', '20180331', '20190331',
                       '20200331', '20210331']

    pair_names = []
    ddays = []
    #get list of conforming pairs and their ddays (usually 24 but sometimes 48)
    for ii in np.arange(len(start_dates)):
        start_dt = datetime.strptime(start_dates[ii], '%Y%m%d')
        end_dt = datetime.strptime(end_dates[ii], '%Y%m%d')
        swe_fnames_all = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe')
        pair_names_all = [os.path.basename(swe_fname)[0:17] for swe_fname in swe_fnames_all]
        pair_names_all.sort()

        for pair_name in pair_names_all:
            master_dt = datetime.strptime((os.path.basename(pair_name[0:8])).split('.')[0], '%Y%m%d')
            slave_dt = datetime.strptime((os.path.basename(pair_name[9:17])).split('.')[0], '%Y%m%d')
            if (master_dt > start_dt) and (slave_dt < end_dt):
                pass
                pair_names.append(pair_name)
                ddays.append(int(np.round((slave_dt - master_dt).total_seconds() / 86400)))

    n_pairs = len(pair_names)
    extent = np.asarray(
        [0, par['range_pixel_spacing'] * par['range_samples'] / np.sin(np.deg2rad(par['incidence_angle'])), 0,
         par['azimuth_pixel_spacing'] * par['azimuth_lines']]) / 1000

    if cummulative:
        swe_render_lim = 200.  # mm
    else:
        swe_render_lim = 60. #mm

    nrow=3; ncol=2
    f, axarr = plt.subplots(nrow, ncol, figsize=(fig_width, 3.4))
    nrow=8; ncol=5
    f, axarr = plt.subplots(nrow, ncol)

    #null all grids
    for ii in np.arange(nrow):
        for jj in np.arange(ncol):
            axarr[ii, jj].axis('off')

    swe_all = np.zeros((n_pairs, par.dim[0], par.dim[1]))

    good_count = np.zeros(par.dim)
    hist = None
    #n_pairs=1
    infill_proportion=np.zeros(n_pairs)
    for kk in np.arange(n_pairs):
        print(kk+1, ' of ', n_pairs, '...')
        ii = kk//ncol
        jj = kk-ii*ncol
        if cummulative:
            swe_est = readBin(swe_est_dir + pair_names[kk][9:] + '.swe', par.dim, 'float32')
        else:
            swe_est = readBin(swe_est_dir + pair_names[kk] + '.diff.swe', par.dim, 'float32')

        swe_sd = readBin(swe_est_dir + pair_names[kk] + '.diff.swe.sim_sd', par.dim,'float32')
        coh = readBin(swe_est_dir + pair_names[kk] + '.diff.swe.coh', par.dim, 'float32')
        nan_mask = (~np.isfinite(swe_est)) #| (~np.isfinite(swe_sd))

        if cummulative:
            bad_mask = nan_mask
        else:
            bad_mask = nan_mask | (swe_est <= swes[4]) | (swe_est >= swes[-4])

        crop = cfg['good_crop']
        infill_proportion[kk] = (np.sum(bad_mask[crop[0]:crop[0]+crop[2], crop[1]: crop[1]+ crop[3]])-np.sum(exclude_mask[crop[0]:crop[0]+crop[2], crop[1]: crop[1]+ crop[3]]))/ \
              (crop[2]*crop[3] -np.sum(exclude_mask[crop[0]:crop[0]+crop[2], crop[1]: crop[1]+ crop[3]]))
        print(infill_proportion[kk])

        swe_est[bad_mask] = np.nan
        swe_all[kk, :,:] = swe_est

        good_count[~bad_mask] +=1

        swe_est[bad_mask] = np.nan

        if cummulative:
            bins_x = np.linspace(-200, 200, 100)
        else:
            bins_x = np.linspace(-40, 60, 50)
        bins_y = np.linspace(0., 1., 50)

        H_kk, xedges, yedges = np.histogram2d((swe_est[~bad_mask].flatten()*1000), coh[~bad_mask].flatten(), bins=[bins_x, bins_y])
        if hist is None:
            hist = H_kk
        else:
            hist+=H_kk

        axarr[ii, jj].imshow(swe_est.T[:,::-1]*1000, vmin=-swe_render_lim, vmax=swe_render_lim, cmap=cmap_rdbu, extent=extent)
        axarr[ii, jj].set_yticks([])
        axarr[ii, jj].set_xticks([])
        axarr[ii, jj].axis('on')
        title_txt = pair_names[kk]
        if ddays[kk] > 24:
            title_txt +='*'
        axarr[ii, jj].text(extent[1]//2, extent[3],title_txt,horizontalalignment='center', verticalalignment='bottom', size=8)

    print('mean infill proportion = ', np.mean(infill_proportion))
    plt.subplots_adjust(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)
    plt.savefig(fig_dir+'swe_maps_all.png')


def make_infill_maps(cfg):
    #fill holes in all swe and swe_sd maps
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    swe_est_dir = cfg['swe_est_dir']
    win_smth = cfg['win']*4
    swe_range = cfg['swe_range']
    swe_del = cfg['swe_del']
    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del

    exclude_mask = get_water_mask()
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    swe_fnames = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe')
    pair_names = [os.path.basename(swe_fname)[0:17] for swe_fname in swe_fnames]
    n_pairs = len(pair_names)
    for jj in np.arange(n_pairs):
        print(jj+1, n_pairs)
        swe_est = readBin(swe_est_dir + pair_names[jj] + '.diff.swe', par.dim,
                          'float32')
        swe_var = readBin(swe_est_dir + pair_names[jj] + '.diff.swe.sim_sd', par.dim,
                          'float32') ** 2

        # a very small number of extra nans in the sim_sd layers is messing things up to synchronize with swe_est...
        swe_var[~np.isfinite(swe_var)] = 0
        swe_var[~np.isfinite(swe_est)] = np.nan

        nan_mask = (~np.isfinite(swe_est))
        bad_mask = nan_mask | (swe_est <= swes[4]) | (swe_est >= swes[-4])

        swe_est[bad_mask] = np.nan
        swe_var[bad_mask] = np.nan

        good_mask = np.zeros(swe_est.shape)
        good_mask[np.isfinite(swe_est)] = 1.
        normalization = uniform_filter(good_mask, win_smth)

        swe_est_zeros = np.copy(swe_est)
        swe_est_zeros[good_mask == 0] = 0
        swe_est_smth = uniform_filter(swe_est_zeros, win_smth) / normalization
        swe_est[bad_mask] = swe_est_smth[bad_mask]
        swe_est[exclude_mask] = np.nan

        swe_var_zeros = np.copy(swe_var)
        swe_var_zeros[good_mask == 0] = 0
        swe_var_smth = uniform_filter(swe_var_zeros, win_smth) / normalization
        swe_var[bad_mask] = swe_var_smth[bad_mask]
        swe_var[exclude_mask] = np.nan

        writeBin(swe_est_dir + pair_names[jj] + '.diff.swe.infill', swe_est)
        writeBin(swe_est_dir + pair_names[jj] + '.diff.swe.sim_sd.infill', swe_var**0.5)


def make_cumulative_maps(cfg):
    infill = True
    #cumulative SWE from Oct 1 to Mar31 each year
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    swe_est_dir = cfg['swe_est_dir']
    win_smth = cfg['win']*4
    swe_range = cfg['swe_range']
    swe_del = cfg['swe_del']
    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del

    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    start_dates = ['20121001', '20131001', '20141001', '20151001', '20161001', '20171001', '20181001',
                   '20191001']
    end_dates = ['20130501', '20140501', '20150501', '20160501', '20170501', '20180501',
                   '20190501', '20200501']

    mean_swe_all = np.zeros(par.dim)
    mean_var_all = np.zeros(par.dim)
    scene_count_all = 0
    for ii in np.arange(len(start_dates)):
        start_date = start_dates[ii]
        end_date = end_dates[ii]
        start_dt = datetime.strptime(start_date, '%Y%m%d')
        end_dt = datetime.strptime(end_date, '%Y%m%d')

        swe_fnames_all = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe')
        pair_names_all = [os.path.basename(swe_fname)[0:17] for swe_fname in swe_fnames_all]
        pair_names_all.sort()
        pair_names = []
        for pair_name in pair_names_all:
            master_dt = datetime.strptime(pair_name[0:8], '%Y%m%d')
            slave_dt = datetime.strptime(pair_name[9:], '%Y%m%d')
            if (master_dt > start_dt) and (master_dt < end_dt) and (slave_dt > start_dt) and (slave_dt < end_dt):
                pair_names.append(pair_name)
        n_pairs = len(pair_names)
        print(start_date, n_pairs)

        extent = np.asarray(
            [0, par['range_pixel_spacing'] * par['range_samples'] / np.sin(np.deg2rad(par['incidence_angle'])), 0,
             par['azimuth_pixel_spacing'] * par['azimuth_lines']]) / 1000
        swe_render_lim = 40. #mm
        swe_sd_render_lim = 30. #mm

        cumulative_swe = np.zeros(par.dim)
        cumulative_var = np.zeros(par.dim)
        for jj in np.arange(n_pairs):
            if infill:
                swe_est = readBin(swe_est_dir + pair_names[jj] + '.diff.swe.infill', par.dim,
                                  'float32')
                swe_var = readBin(swe_est_dir + pair_names[jj] + '.diff.swe.sim_sd.infill', par.dim,
                                  'float32') ** 2
            else:
                swe_est = readBin(swe_est_dir + pair_names[jj] + '.diff.swe', par.dim,
                                  'float32')
                swe_var = readBin(swe_est_dir + pair_names[jj] + '.diff.swe.sim_sd', par.dim,
                                 'float32')**2

                #a very small number of extra nans in the sim_sd layers is messing things up to synchronize with swe_est...
                swe_var[~np.isfinite(swe_var)] = 0
                swe_var[~np.isfinite(swe_est)] = np.nan

            cumulative_swe+=swe_est
            cumulative_var+=swe_var

            mean_swe_all+=swe_est
            mean_var_all+=swe_var
            scene_count_all+=1

            writeBin(swe_est_dir+pair_names[jj][9:]+'.swe', cumulative_swe)
            writeBin(swe_est_dir + pair_names[jj][9:] + '.swe_sd', cumulative_var**0.5)

    mean_swe_all/=scene_count_all
    mean_var_all/= scene_count_all
    writeBin(swe_est_dir + 'mean_swe_all', mean_swe_all)
    writeBin(swe_est_dir + 'mean_sd_all', mean_var_all ** 0.5)

def plot_mean_swe_all(cfg):
    #plt.global mean of
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    swe_est_dir = cfg['swe_est_dir']
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    extent = np.asarray(
        [0, par['range_pixel_spacing'] * par['range_samples'] / np.sin(np.deg2rad(par['incidence_angle'])), 0,
         par['azimuth_pixel_spacing'] * par['azimuth_lines']]) / 1000
    swe_render_lim = 40.  # mm

    mean_swe_all = readBin(swe_est_dir + 'mean_swe_all', par.dim,'float32')

    plt.figure(figsize=(fig_width, 2.5))
    plt.imshow(mean_swe_all.T[:, ::-1] * 1000, vmin=-swe_render_lim, vmax=swe_render_lim, cmap=cmap_rdbu,extent=extent)
    plt.gca().set_yticks([])
    plt.gca().set_xticks([])
    plt.tight_layout()
    plt.savefig(fig_dir+'mean_swe_all.png')

def get_sarpix(lon, lat):
    stack = stack_SLA27D
    looks_str = '3_12'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    dem_fname = stack._dir + 'dem_'+looks_str+'/seg.dem'
    dem_par = DEM_Par(stack._dir + 'dem_'+looks_str+'/seg.dem_par')
    dem = DEM(stack._dir + 'dem_'+looks_str+'/seg.dem', par = dem_par)
    hgt = dem.interpolate_map_coords([lon+1j*lat])

    result = run('coord_to_sarpix', par, None, None, lat, lon, hgt)

    for line in result.stdout.split('\n'):
        if 'SLC/MLI range, azimuth pixel (int):' in line:
            toks = line.split(':')[1].split(' ')
            toks = [t for t in toks if t]
            # RG/Az
            rg_idx = int(toks[0])
            az_idx = int(toks[1])
            return rg_idx, az_idx

def insitu_rawdata_plot():
    fig = plt.figure(figsize=(fig_width-0.5, 6))
    transects = [t_sa_2018, t_sb_2018, t_sc_2018, t_ensom_2018] + transects_enr
    altcolor = 'tab:green'

    for ii in np.arange(len(transects)):
        plt.subplot(8,1,ii+1)
        ax1 = plt.gca()
        plt.plot(np.arange(len(transects[ii].distance))+1, transects[ii].depth*1000., color='k', marker = 'o',
                 linestyle = '-', fillstyle='none')

        #plt.xlim((0,200.))
        plt.xlim((1, 21))
        plt.ylim((0, 1000.))
        if ii + 1 == 8:
            plt.xlabel('Transect sample #')
            plt.xticks([0,5,10,15,20])
        else:
            plt.xticks([])
            plt.xlabel('\n')
        ax2 = plt.gca().twinx()
        density_ii = transects[ii].swe / transects[ii].depth
        ax2.plot(np.arange(len(transects[ii].distance))+1, density_ii, color=altcolor, marker = 's',
                 linestyle = '-', fillstyle='none')
        #ax2.set_ylabel('Relative snow density')
        ax2.set_xlim((1, 21))
        ax2.set_ylim((0, 0.4))
        ax2.set_yticks([0, 0.2, 0.4])
        ax2.tick_params(axis='y', labelcolor=altcolor)

        print('Transect:', ii+1,', mean depth:', np.mean(transects[ii].depth*1000.), ', mean density:',np.mean(density_ii) )

    #plt.tight_layout()
    plt.savefig(fig_dir + 'insitu_rawdata.png')

def insitu_scatterplot(cfg):
    fig = plt.figure(figsize=(fig_width, 3))
    err_arr = []
    print('Eppler_Morse_2018_transects:')
    # transects = [t_sa_2018, t_sb_2018, t_sc_2018]
    # for transect in transects:
    err_arr+=[plot_validation_scatter_point(cfg, 0.5*(t_sa_2018.start_pos + t_sa_2018.end_pos), '20180315', np.nanmean(t_sa_2018.swe), insitu_swesd=(np.nanvar(t_sa_2018.swe)/np.sum(np.isfinite(t_sa_2018.swe)))**0.5, marker = 'o', label='A - 2018-03-15')]
    err_arr+=[plot_validation_scatter_point(cfg, 0.5*(t_sb_2018.start_pos + t_sb_2018.end_pos), '20180315', np.nanmean(t_sb_2018.swe), insitu_swesd=(np.nanvar(t_sb_2018.swe)/np.sum(np.isfinite(t_sb_2018.swe)))**0.5, marker = 's', label='B - 2018-03-15')]
    err_arr+=[plot_validation_scatter_point(cfg, 0.5*(t_sc_2018.start_pos + t_sc_2018.end_pos), '20180315', np.nanmean(t_sc_2018.swe), insitu_swesd=(np.nanvar(t_sc_2018.swe)/np.sum(np.isfinite(t_sc_2018.swe)))**0.5, marker = 'D', label='C - 2018-03-15')]

    #Tim Ensom transects
    print('Ensom_transects:')
    ensom_pos1 = utm_to_geo(561164.2, 7578872.5, 8) #Tim Ensom 2018 survey (20180422)
    err_arr += [plot_validation_scatter_point(cfg, ensom_pos1, '20180422', 0.107, insitu_swesd=0.0057, marker = '*', label='D - 2018-04-22')]

    #ENR transects
    print('ENR_transects:')
    enr_pos = (-133.532494, 68.318757) #ENR surveys at Inuvik Satellite facility
    err_arr+=[plot_validation_scatter_point(cfg, enr_pos, '20170201', 0.0590, insitu_swesd=0.003785939, marker = 'v', label='E - 2017-02-01')]
    err_arr+=[plot_validation_scatter_point(cfg, enr_pos, '20170215', 0.0985, insitu_swesd=0.00799479, marker = '^', label='E - 2017-02-15')]
    err_arr+=[plot_validation_scatter_point(cfg, enr_pos, '20170301', 0.0920, insitu_swesd=0.005537749, marker = '<', label='E - 2017-03-01')]
    err_arr+=[plot_validation_scatter_point(cfg, enr_pos, '20170316', 0.1200, insitu_swesd=0.01122992, marker = '>', label='E - 2017-03-16')]

    plt.plot([0, 200], [0, 200], 'k', label='1:1')
    plt.xlim((0., 150))
    plt.ylim((0., 200))
    plt.title('Comparison with In Situ Measurements')
    plt.xlabel('In situ measured SWE [mm]')
    plt.ylabel('SlopeVar estimator\ncumulative SWE [mm]')
    # get handles
    handles, labels = plt.gca().get_legend_handles_labels()
    # remove the errorbars
    handles_new = [h[0] for h in handles[1:]]
    handles = handles_new+[handles[0]]
    labels = labels[1:] + [labels[0]]
    # use them in the legend
    plt.gca().legend(handles, labels, loc='upper left', numpoints=1, prop={'size': 6})

    plt.tight_layout()
    plt.savefig(fig_dir + 'insitu_scatter.png')

    err_arr = np.asarray(err_arr)
    global_rmse = np.mean(err_arr**2)**0.5
    global_bias = np.mean(err_arr)

    print('Global RMSE [mm]: ', global_rmse*1000.)
    print('Global bias [mm]: ', global_bias*1000.)
    x=1

def plot_validation_scatter_point(cfg, pos, plot_date, insitu_swe, insitu_swesd=None, label=None, marker=None):
    #pos is [lon, lat]
    print(label, ':', pos)
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    win = cfg['win']
    swe_est_dir = cfg['swe_est_dir']
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    n_sd = 1 #number of sds for error bars
    xi_all = get_xi(cfg)
    exclude_mask = get_water_mask(cfg) #just exclude water areas since the blunders were masked out and filled already
    xi_all[exclude_mask]=np.nan

    rg_idx, az_idx = get_sarpix(pos[0], pos[1])
    swe_fname_list = glob.glob(swe_est_dir+'????????.swe')
    swe_fname_list.sort()

    #find closest in time swe maps that bracket
    dates_list = [os.path.basename(fname)[0:8] for fname in swe_fname_list]
    dt_list = [datetime.strptime(date, '%Y%m%d') for date in dates_list]
    plot_dt = datetime.strptime(plot_date, '%Y%m%d')
    dt_diff = np.asarray([(dt_list_-plot_dt).total_seconds() for dt_list_ in dt_list])
    idx_nearest = np.argmin(np.abs(dt_diff))
    if dt_list[idx_nearest] < plot_dt:
        idx_pre = idx_nearest
    else:
        idx_pre = idx_nearest-1
    idx_post = idx_pre+1
    swe_pre_all = readBin(swe_fname_list[idx_pre], par.dim, 'float32')
    swe_post_all = readBin(swe_fname_list[idx_post], par.dim, 'float32')
    swe_pre = swe_pre_all[rg_idx, az_idx]
    swe_post = swe_post_all[rg_idx, az_idx]

    xi_win = xi_all[rg_idx-win[0]//2: rg_idx+win[0]//2, az_idx-win[1]//2: az_idx+win[1]//2]
    xi_sd = np.nanstd(xi_win)
    good_count = np.sum(np.isfinite(xi_win))

    # get 'roughness' DEM standard deviation
    dem_all = get_dem(cfg)
    dem_win = dem_all[rg_idx - win[0] // 2: rg_idx + win[0] // 2, az_idx - win[1] // 2: az_idx + win[1] // 2]
    x_idx = np.outer(np.arange(win[0]), np.ones(win[1]))
    y_idx = np.outer(np.ones(win[0]), np.arange(win[1]))
    poly = lsq2_fit(dem_win.flatten(),x_idx.flatten(), y_idx.flatten(),1)[0]
    dem_fit = poly2(x_idx, y_idx, poly,1)[0]
    dem_sd = np.std(dem_win-dem_fit)

    dt_fraction = (plot_dt - dt_list[idx_pre]).total_seconds()/(dt_list[idx_post] - dt_list[idx_pre]).total_seconds()
    print(good_count, win[0]*win[1], xi_sd, dem_sd)
    swe_interp = (1-dt_fraction)*swe_pre + dt_fraction*swe_post

    swesd_pre_all = readBin(swe_fname_list[idx_pre]+'_sd', par.dim, 'float32')
    swesd_post_all = readBin(swe_fname_list[idx_post]+'_sd', par.dim, 'float32')
    swesd_pre = swesd_pre_all[rg_idx, az_idx]
    swesd_post = swesd_post_all[rg_idx, az_idx]
    swesd_interp = ((1-dt_fraction)*swesd_pre**2 + dt_fraction*swesd_post**2)**0.5
    yerr = n_sd*swesd_interp * 1000
    xerr=None
    if insitu_swesd is not None:
        xerr = n_sd*insitu_swesd*1000

    plt.errorbar([insitu_swe * 1000], [swe_interp*1000], xerr=xerr, yerr=yerr, capsize=2, label=label, marker=marker, linestyle="None")
    return swe_interp - insitu_swe #returns the 'error'

def get_nalcms_agg(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    dem_dir = stack._dir + 'dem_' + looks_str + '/'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')

    nalcms = readBin(dem_dir + 'topo/nalcms2015.rdc', par.dim, np.ubyte)
    #create an aggragated version of nalcms
    # nalcms_agg_map = {1:0, 2:0, 5:0, 6:0, 8:1, 11:1, 10:2, 12:2, 13:2, 16:2, 14:3, 17:4, 18:5}
    # nalcms_agg_names = ['forest', 'shrub', 'barren/grass', 'wetland', 'urban', 'water']

    nalcms_agg_map = {1:0, 2:0, 5:1, 6:1, 8:2, 11:2, 10:3, 12:3, 13:3, 16:3, 14:4, 17:5, 18:6}
    nalcms_agg_names = ['Needleleaf forest', 'Deciduous/mixed forest', 'Shrub', 'Grassland/barren', 'Wetland', 'Urban', 'Water']
    nalcms_agg = np.zeros_like(nalcms)
    for key in nalcms_agg_map:
        nalcms_agg[nalcms == key] = nalcms_agg_map[key]

    return nalcms_agg_names, nalcms_agg


def temporal_validation_plot(cfg, xi_sd_thresh = None):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    swe_est_dir = cfg['swe_est_dir']
    swe_range = cfg['swe_range']
    swe_del = cfg['swe_del']

    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    #inc = readBin(stack._dir + 'dem_' + looks_str + '/topo/inc.rdc', par.dim, 'float32')
    #DO NOT use local incidence angle - want horizontal projection
    #TODO re-run this and edit paper tables, plots and text to match!!!!!!!
    inc = np.ones((par.dim)) * np.deg2rad(par['incidence_angle'])  # just use scene centre incidence everywhere
    ground_range_projection_factor = 1/np.sin(inc) #scaling factor to go from slant range to ground range

    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del
    swe_fnames_all = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe')
    swe_fnames_all.sort()
    swe_fnames_all =swe_fnames_all

    xi_sd = get_xi_sd(cfg)

    n_swe = len(swe_fnames_all)

    dict_daily = parse_ec_dir('./inuvik/met_data/inuvik_climate/', 'daily', plot=False)
    era5 = get_inuvik_RS2_SLA27D_era5()

    #now go through and assign values to regular 24 day grid

    date_start = os.path.basename(swe_fnames_all[0])[0:8]
    date_end = os.path.basename(swe_fnames_all[-1])[0:8]
    datetime_start = datetime.strptime(date_start, '%Y%m%d')
    datetime_end = datetime.strptime(date_end, '%Y%m%d')
    n_intervals = int(np.round((datetime_end-datetime_start).total_seconds()/86400/24)+1)

    d_swe_era5 = np.zeros(n_intervals)
    d_swe_era5[:] = np.nan

    datetime_mid = []
    no_snow_mask = np.zeros(n_intervals, dtype=bool)
    for ii in np.arange(n_intervals):
        datetime_m = datetime_start + timedelta(days=ii*24.)
        datetime_s = datetime_m + timedelta(days= 24)
        date_m = datetime_m.strftime('%Y%m%d')
        date_s = datetime_s.strftime('%Y%m%d')
        datetime_mid = datetime_mid + [datetime_m + 0.5*(datetime_s-datetime_m)]

        d_swe_era5[ii] = era5['sd'][era5['times'].index(datetime_s)] - era5['sd'][era5['times'].index(datetime_m)]
        if (era5['sd'][era5['times'].index(datetime_s)] == 0) and (era5['sd'][era5['times'].index(datetime_m)] == 0):
            no_snow_mask[ii] = True

    if True:
        mean_swe = np.zeros(n_intervals)
        mean_swe[:] = np.nan
        for ii in np.arange(n_swe):
            swe_fname = swe_fnames_all[ii]
            swe_est = readBin(swe_fname, par.dim, 'float32')

            basename = os.path.basename(swe_fname)
            date_m = basename[0:8]
            date_s = basename[9:17]
            datetime_m = datetime.strptime(date_m, '%Y%m%d')
            datetime_s = datetime.strptime(date_s, '%Y%m%d')
            d_days = int(np.round((datetime_s - datetime_m).total_seconds() / 86400.))

            idx = int(np.round((datetime_m - datetime_start).total_seconds()/86400/24))
            swe_render_lim = np.max(np.abs(swe_range))
            bad_mask = (swe_est.flatten() <= swes[4]) | (swe_est.flatten() >= swes[-4]) | ~np.isfinite(swe_est.flatten())
            if xi_sd_thresh is not None:
                bad_mask = bad_mask | (xi_sd < xi_sd_thresh).flatten()
            #compute area means, factoring in a slant-to-ground range scaling factor
            mean_swe_ii = np.mean(((swe_est*ground_range_projection_factor).flatten()[~bad_mask]))/np.mean(((ground_range_projection_factor).flatten()[~bad_mask]))
            if d_days == 24:
                mean_swe[idx] = mean_swe_ii
            elif d_days == 48:
                mean_swe[idx] = mean_swe_ii/2
                mean_swe[idx+1] = mean_swe_ii/2
            else:
                raise ValueError('unexpected d_days')

            print(ii,basename[0:17])
        mean_swe.tofile(swe_est_dir+'temporal_validation_plot.mean_swe.save')
    else:
        mean_swe = np.fromfile(swe_est_dir+'temporal_validation_plot.mean_swe.save')

    doy = np.asarray([datetime_mid_.timetuple().tm_yday for datetime_mid_ in datetime_mid])
    #compute RMSE, bias and pearson correlation coefficents
    all_mask = np.isfinite(mean_swe)
    q1_mask = (doy < 0.25*365.25) & all_mask
    q2_mask = (doy >= 0.25 * 365.25) & (doy < 0.5 * 365.25) & all_mask
    q3_mask = (doy >= 0.5 * 365.25) & (doy < 0.75 * 365.25) & all_mask
    q4_mask = (doy >= 0.75 * 365.25)  & all_mask
    q14_mask = q1_mask | q4_mask
    no_snow_mask = no_snow_mask & all_mask# & q3_mask

    def compute_stats(s0, s1, label=''):
        bias = np.mean(s1 - s0)
        rmse = np.mean((s1-s0)**2)**0.5

        s0_dm = s0 - np.mean(s0)
        s1_dm = s1 - np.mean(s1)
        corr = np.mean(s0_dm*s1_dm)/(np.mean(s0_dm*s0_dm)*np.mean(s1_dm*s1_dm))**0.5
        print('{0}: {1} {2:.1f} {3:.1f} {4:.2f}'.format(label, len(s0), bias*1000, rmse*1000, corr))
        return bias, rmse, corr

    print('temporal comparison stats wrt ERA5')
    print('Case, number, bias [mm], RMSE [mm], Correlation')

    bias_q1, rmse_q1, corr_q1 = compute_stats(d_swe_era5[q1_mask], mean_swe[q1_mask], label = 'Jan-Mar')
    bias_q2, rmse_q2, corr_q2 = compute_stats(d_swe_era5[q2_mask], mean_swe[q2_mask], label = 'Apr-Jun')
    bias_q3, rmse_q3, corr_q3 = compute_stats(d_swe_era5[q3_mask], mean_swe[q3_mask], label = 'Jul-Sep')
    bias_q4, rmse_q4, corr_q4 = compute_stats(d_swe_era5[q4_mask], mean_swe[q4_mask], label = 'Oct-Dec')
    bias_all, rmse_all, corr_all = compute_stats(d_swe_era5[all_mask], mean_swe[all_mask], label = 'All')
    bias_q14, rmse_q14, corr_q14 = compute_stats(d_swe_era5[q14_mask], mean_swe[q14_mask], label = 'Oct-Mar')
    bias_no_snow, rmse_no_snow, corr_no_snow = compute_stats(d_swe_era5[no_snow_mask], mean_swe[no_snow_mask], label='no_snow')

    plt.figure(figsize=(fig_width_full, 5.5))

    plt.subplot(2,1,1)
    plt.plot(datetime_mid, d_swe_era5 * 1000,':', color='gray', label='ERA5')

    for ii in np.arange(n_intervals):
        marker = '*'
        color= 'blue'
        if doy[ii]>0.25*365.25:
            marker = 'o'
            color = 'orange'
        if doy[ii]>0.5*365.25:
            marker = '^'
            color = 'green'
        if doy[ii] > 0.75*365.25:
            marker = '+'
            color = 'cyan'
        plt.plot([datetime_mid[ii]], [mean_swe[ii] * 1000], marker=marker, color=color)

    plt.plot([datetime_mid[0]], [-100], marker='*', color='blue', label='SlopeVar estimator (Jan-Mar)', linestyle='None')
    plt.plot([datetime_mid[0]], [-100], marker='o', color='orange', label='SlopeVar estimator (Apr-Jun)', linestyle='None')
    plt.plot([datetime_mid[0]], [-100], marker='^', color='green', label='SlopeVar estimator (Jul-Sep)', linestyle='None')
    plt.plot([datetime_mid[0]], [-100], marker='+', color='cyan', label='SlopeVar estimator (Oct-Dec)', linestyle='None')
    plt.ylim((-10,60))
    plt.title('Time-series comparison with ERA5')
    plt.xlabel('Date')
    plt.ylabel(r'24 day $\Delta$SWE [mm]')
    plt.legend(loc='upper left')

    plt.subplot(2,1, 2)

    for ii in np.arange(n_intervals):
        marker = '*'
        color= 'blue'
        if doy[ii]>0.25*365.25:
            marker = 'o'
            color = 'orange'
        if doy[ii]>0.5*365.25:
            marker = '^'
            color = 'green'
        if doy[ii] > 0.75*365.25:
            marker = '+'
            color = 'cyan'
        plt.plot([d_swe_era5[ii] * 1000], [mean_swe[ii] * 1000], marker=marker, color=color)
    plt.plot([-100], [-100], marker='*', color='blue', label='Jan-Mar', linestyle='None')
    plt.plot([-100], [-100], marker='o', color='orange', label='Apr-Jun', linestyle='None')
    plt.plot([-100], [-100], marker='^', color='green', label='Jul-Sep', linestyle='None')
    plt.plot([-100], [-100], marker='+', color='cyan', label='Oct-Dec', linestyle='None')
    plt.plot([-100, 100], [-100, 100], 'k', label='1:1')

    plt.xlim([-90, 40])
    plt.ylim([-10, 40])
    plt.title('Scatter-plot comparison with ERA5')
    plt.xlabel(r'24 day $\Delta$SWE (ERA5) [mm]')
    plt.ylabel(r'24 day $\Delta$SWE' '\n(SlopeVar estimator) [mm]')
    plt.legend()
    plt.tight_layout()
    plt.savefig(fig_dir + 'temporal_validation.png')

def blown_snow_analysis():
    stack = stack_SLA27D
    #generate plot of fractional swe estimation error caused by wrong density
    par = SLC_Par(stack._dir + 'rslc/master.rslc.par')
    fractional_day = 0.63
    era5_data = get_inuvik_RS2_SLA27D_era5()
    dates = ['20120105','20120129', '20120222', '20120317', '20120410', '20120504', '20120528', '20120621', '20120715', '20120808',
                 '20120901', '20120925', '20121019', '20121112', '20121206', '20121230', '20130123', '20130216',
                 '20130312', '20130405', '20130429', '20130523', '20130616', '20130710', '20130803', '20130827',
                 '20130920', '20131014', '20131107', '20131225', '20140118', '20140424', '20140518', '20140611',
                 '20140705', '20140729', '20140822', '20140915', '20141009', '20141102', '20141126', '20141220',
                 '20150113', '20150206', '20150302', '20150513', '20150606', '20150724', '20150817', '20150910',
                 '20151004', '20151028', '20151121', '20151215', '20160108', '20160225', '20160320', '20160624',
                 '20160718', '20160811', '20160904', '20160928', '20161022', '20161115', '20161209', '20170102',
                 '20170126', '20170219', '20170315', '20170408', '20170502', '20170526', '20170619', '20170713',
                 '20170806', '20170830', '20170923', '20171017', '20171204', '20171228', '20180214', '20180310',
                 '20180403', '20180427', '20180521', '20180614', '20180708', '20180801', '20180825', '20180918',
                 '20181012', '20181105', '20181129', '20181223', '20190116', '20190209', '20190305', '20190329',
                 '20190422', '20190516', '20190609', '20190703', '20190727', '20190913', '20191007', '20191124',
                 '20200111', '20200204', '20200228', '20200323', '20200416']


    #replace bsi with version2
    era5_data['bsi_old'] = era5_data['bsi']
    era5_data['bsi'] = snow.blown_snow_index2(era5_data)*24

    dts = [datetime.strptime(date, '%Y%m%d') + timedelta(fractional_day) for date in dates]

    era5_data_pairwise = snow.get_era5_pairwise_data(era5_data, dates, fractional_day, plot=False)

    plt.figure(figsize=(fig_width, 3))

    plt.polar(np.diagonal(era5_data_pairwise['cum_bsi_vec_dir'], offset=1)[0] / 180 * np.pi,
              np.diagonal(era5_data_pairwise['cum_bsi_vec_mag'], offset=1)[0], '*', label='20120105_20120129')
    plt.polar(np.diagonal(era5_data_pairwise['cum_bsi_vec_dir'], offset=1)[1:]/180*np.pi, np.diagonal(era5_data_pairwise['cum_bsi_vec_mag'],
                offset=1)[1:], '*', label='all other 24-day intervals')
    plt.gca().set_theta_direction(-1)
    plt.gca().set_theta_zero_location('N')
    los_heading = np.exp(1j*np.deg2rad(par['heading']) + np.pi/2)
    plt.polar((los_heading, los_heading+np.pi), (60, 60), '--',label = 'SAR look direction axis')
    plt.legend(loc='lower right', frameon=True)
    plt.tight_layout()
    plt.savefig(fig_dir + 'bsi.png')
    x=1

def delta_k_comparison(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    win = cfg['win']

    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    dim = par.dim
    xi = get_xi(cfg)
    xi_mean = uniform_filter(xi, win)
    xi_sd = np.sqrt(uniform_filter(xi ** 2, win) - xi_mean ** 2)

    exclude_mask = get_water_mask(cfg)  # just exclude water areas since the blunders were masked out and filled already

    B = par['chirp_bandwidth'] / par['radar_frequency']
    beta = 1. / 3

    sigma_ratio = 2**0.5 * xi_sd / (B * (1 - beta) * beta ** 0.5 * xi_mean)
    print('ratio median = ',np.median(sigma_ratio[~exclude_mask].flatten()))
    print('ratio <1: ', np.sum(sigma_ratio[~exclude_mask].flatten()<1)/len(sigma_ratio[~exclude_mask].flatten()))

    plt.figure(figsize=(fig_width, 2))
    hist_log, bin_edges = np.histogram(np.log10(sigma_ratio[~exclude_mask].flatten()), bins=200, range=(-1,2))
    plt.bar(10**((bin_edges[0:-1] + bin_edges[1:])/2), hist_log, width = 10**bin_edges[1:]-10**bin_edges[0:-1])
    y_max = plt.gca().get_ylim()[1]
    plt.plot([1,1],[0, y_max], color='gray')
    plt.ylim((0, y_max))
    plt.xlim((0.5, 30))
    plt.gca().set_xscale('log')

    plt.gca().axes.yaxis.set_ticks((0, y_max))
    plt.gca().set_yticklabels(['0', 'max'])
    plt.xlabel(r'Precision ratio: $\sigma_{\Delta{S\_DK}}/\sigma_{\Delta{S\_SV}}$')
    plt.ylabel('Frequency')
    plt.tight_layout()
    plt.savefig(fig_dir + 'delta_k_comparison.png')

def get_water_mask(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    water_mask_fname = stack._dir + 'rmli_' + looks_str + '/water_mask_' + looks_str + '.ras'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    water_mask = read_ras(water_mask_fname)[0].T == 0
    return water_mask

def get_xi(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    topo_dir = stack._dir + 'dem_' + looks_str + '/topo/'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    xi = readBin(topo_dir + 'dphi_dswe.rdc', par.dim, 'float32').astype(np.float32)
    return xi

def get_xi_sd(cfg):
    # get standard deviation of xi
    win = cfg['win']
    exclude_mask = get_water_mask(cfg)
    xi = get_xi(cfg)
    xi_masked = np.copy(xi)
    xi_masked[exclude_mask] = 0
    good_mask = np.ones_like(xi)
    good_mask[exclude_mask] = 0
    good_mask_filtered = uniform_filter(good_mask, win)
    xi_sd = np.sqrt(uniform_filter(xi_masked ** 2, win) / good_mask_filtered - (
                uniform_filter(xi_masked, win) / good_mask_filtered) ** 2)
    return xi_sd

def get_dem(cfg):
    # get heights
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    dem_dir = stack._dir + 'dem_' + looks_str + '/'
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    dem = readBin(dem_dir + 'seg.dem.rdc', par.dim, 'float32')
    return dem

    win = cfg['win']
    exclude_mask = get_water_mask(cfg)
    xi = get_xi(cfg)
    xi_masked = np.copy(xi)
    xi_masked[exclude_mask] = 0
    good_mask = np.ones_like(xi)
    good_mask[exclude_mask] = 0
    good_mask_filtered = uniform_filter(good_mask, win)
    xi_sd = np.sqrt(uniform_filter(xi_masked ** 2, win) / good_mask_filtered - (
                uniform_filter(xi_masked, win) / good_mask_filtered) ** 2)
    return xi_sd

def get_seasonal_masks(cfg):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    swe_est_dir = cfg['swe_est_dir']
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')

    swe_fnames_all = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe.infill')
    swe_fnames_all.sort()

    n_swe = len(swe_fnames_all)
    era5 = get_inuvik_RS2_SLA27D_era5()

    date_start = os.path.basename(swe_fnames_all[0])[0:8]
    date_end = os.path.basename(swe_fnames_all[-1])[0:8]
    datetime_start = datetime.strptime(date_start, '%Y%m%d')
    datetime_end = datetime.strptime(date_end, '%Y%m%d')
    n_intervals = int(np.round((datetime_end-datetime_start).total_seconds()/86400/24)+1)

    d_swe_era5 = np.zeros(n_intervals)
    d_swe_era5[:] = np.nan

    datetime_mid = []
    no_snow_mask = np.zeros(n_intervals, dtype=bool)
    for ii in np.arange(n_intervals):
        datetime_m = datetime_start + timedelta(days=ii*24.)
        datetime_s = datetime_m + timedelta(days= 24)
        datetime_mid = datetime_mid + [datetime_m + 0.5*(datetime_s-datetime_m)]

        d_swe_era5[ii] = era5['sd'][era5['times'].index(datetime_s)] - era5['sd'][era5['times'].index(datetime_m)]
        if (era5['sd'][era5['times'].index(datetime_s)] == 0) and (era5['sd'][era5['times'].index(datetime_m)] == 0):
            no_snow_mask[ii] = True

    doy = np.asarray([datetime_mid_.timetuple().tm_yday for datetime_mid_ in datetime_mid])
    #compute RMSE, bias and pearson correlation coefficents
    all_mask = np.ones(n_intervals, dtype=bool)

    masks = {}
    masks['data'] = np.zeros(n_intervals, dtype=bool)
    masks['q1'] = (doy < 0.25*365.25) & all_mask
    masks['q2'] = (doy >= 0.25 * 365.25) & (doy < 0.5 * 365.25) & all_mask
    masks['q3'] = (doy >= 0.5 * 365.25) & (doy < 0.75 * 365.25) & all_mask
    masks['q4'] = (doy >= 0.75 * 365.25)  & all_mask
    masks['q14'] = masks['q1'] | masks['q4']
    masks['no_snow'] = no_snow_mask & all_mask

    for ii in np.arange(n_swe):
        swe_fname = swe_fnames_all[ii]
        basename = os.path.basename(swe_fname)
        date_m = basename[0:8]
        date_s = basename[9:17]
        datetime_m = datetime.strptime(date_m, '%Y%m%d')
        datetime_s = datetime.strptime(date_s, '%Y%m%d')
        idx = int(np.round((datetime_m - datetime_start).total_seconds() / 86400 / 24))
        masks['data'][idx] = True

    return masks

def seasonal_mean_maps(cfg, thresh_xi_sd = (0., 1e6)):
    stack = cfg['stack']
    looks_str = cfg['looks_str']
    swe_est_dir = cfg['swe_est_dir']
    swe_range = cfg['swe_range']
    swe_del = cfg['swe_del']
    par = MLI_Par(stack._dir + 'rmli_' + looks_str + '/rmli_' + looks_str + '.ave.par')
    exclude_mask = get_water_mask(cfg)
    xi_sd = get_xi_sd(cfg)
    xi_sd[exclude_mask] = np.nan
    print(np.nanpercentile(xi_sd.flatten(),25.))

    swes = swe_range[0] + np.arange(int((swe_range[1] - swe_range[0]) / swe_del)) * swe_del
    swe_fnames_all = glob.glob(swe_est_dir + '????????_????????' + '.diff.swe.infill')

    swe_fnames_all.sort()

    n_swe = len(swe_fnames_all)

    date_start = os.path.basename(swe_fnames_all[0])[0:8]
    date_end = os.path.basename(swe_fnames_all[-1])[0:8]
    datetime_start = datetime.strptime(date_start, '%Y%m%d')
    datetime_end = datetime.strptime(date_end, '%Y%m%d')
    n_intervals = int(np.round((datetime_end-datetime_start).total_seconds()/86400/24)+1)

    #compute RMSE, bias and pearson correlation coefficents
    masks = get_seasonal_masks(cfg)

    nbins=100
    bins = np.linspace(-50, 80, nbins + 1)
    def _get_mean_dswe(masks, thresh_xi_sd, bins):
        mean_dswe = {}
        hist_dswe = {}
        counts = {}
        rmse = {}
        root_mean_var = {}
        for key in masks:
            mean_dswe[key] = np.zeros(par.dim)
            hist_dswe[key] = np.zeros(nbins)
            counts[key] = 0
            rmse[key] = 0
            root_mean_var[key] = 0

        mean_swe = np.zeros(n_intervals)
        mean_swe[:] = np.nan
        for ii in np.arange(n_swe):
            swe_fname = swe_fnames_all[ii]
            basename = os.path.basename(swe_fname)
            print(ii, basename[0:17])
            swe_est = readBin(swe_fname, par.dim, 'float32')
            date_m = basename[0:8]
            date_s = basename[9:17]
            datetime_m = datetime.strptime(date_m, '%Y%m%d')
            datetime_s = datetime.strptime(date_s, '%Y%m%d')
            d_days = int(np.round((datetime_s - datetime_m).total_seconds() / 86400.))

            idx = int(np.round((datetime_m - datetime_start).total_seconds()/86400/24))

            nan_mask = (~np.isfinite(swe_est))
            good_mask = np.isfinite(mean_dswe['no_snow'].flatten()) & (xi_sd.flatten() > thresh_xi_sd[0]) & (
                        xi_sd.flatten() < thresh_xi_sd[1])


            bad_mask = nan_mask | (swe_est <= swes[4]) | (swe_est >= swes[-4]) | (xi_sd < thresh_xi_sd[0]) | (
                        xi_sd > thresh_xi_sd[1])

            mse = np.mean(swe_est[~bad_mask]**2)
            var = np.mean((swe_est[~bad_mask] - np.mean(swe_est[~bad_mask]))**2)
            if d_days != 24:
                swe_est /= d_days//24

            for key in masks:
                if masks[key][idx]:
                    mean_dswe[key] +=swe_est
                    counts[key] +=1
                    rmse[key] += mse
                    root_mean_var[key] += var

                    H_kk, _ = np.histogram((swe_est[~bad_mask].flatten() * 1000), bins=bins)
                    hist_dswe[key] += H_kk

        for key in masks:
            mean_dswe[key]/=counts[key]
            rmse[key] = (rmse[key]/counts[key])**0.5
            root_mean_var[key]= (root_mean_var[key]/counts[key])**0.5
        return mean_dswe, rmse, root_mean_var, hist_dswe
    #####end
    mean_dswe, rmse, root_mean_var, hist_dswe = _get_mean_dswe( masks, thresh_xi_sd, bins)
    _, _, _, hist_dswe_thresh = _get_mean_dswe(masks, (7.585708141326904, 1e6), bins)

    plot_keys = ['q4', 'q1', 'no_snow']
    print('\n***RMSE [mm]***')
    for key in plot_keys:
        print(key, ': ', rmse[key]*1000.)

    print('***root mean var [mm]***')
    for key in plot_keys:
        print(key, ': ', root_mean_var[key]*1000.)

    good_mask = np.isfinite(mean_dswe['no_snow'].flatten()) & (xi_sd.flatten() > thresh_xi_sd[0]) & (xi_sd.flatten() < thresh_xi_sd[1])

    extent = np.asarray(
        [0, par['range_pixel_spacing'] * par['range_samples'] / np.sin(np.deg2rad(par['incidence_angle'])), 0,
         par['azimuth_pixel_spacing'] * par['azimuth_lines']]) / 1000

    plt.figure(figsize=(fig_width/2, 4.2))
    swe_render_lim = 40.
    title_dict = {'q1': 'Jan-Mar', 'q2': 'Apr-Jun', 'q3':'Jul-Sep', 'q4': 'Oct-Dec', 'q14':'Oct-Mar', 'no_snow': 'Snow Free'}
    for ii in np.arange(len(plot_keys)):
        key = plot_keys[ii]
        plt.subplot(3,1,ii+1)
        plt.title(title_dict[key])
        mean_dswe[key][np.reshape(~good_mask, mean_dswe[key].shape)] = np.nan
        plt.imshow(mean_dswe[key].T[:, ::-1] * 1000, vmin=-swe_render_lim, vmax=swe_render_lim, cmap=cmap_rdbu,
               extent=extent)
        plt.yticks([])
        plt.xticks([])
        plt.subplots_adjust(wspace=0, hspace=0.2, left=0, right=1, bottom=0, top=0.95)
        plt.savefig(fig_dir + 'seasonal_distributions_a.png')

    plt.figure(figsize=(fig_width/2, 4.2))
    for ii in np.arange(len(plot_keys)):
        key = plot_keys[ii]

        plt.subplot(3, 1, ii + 1)
        blue = 'tab:blue'
        red = 'tab:red'
        plt.plot(0.5 * (bins[0:-1] + bins[1:]), hist_dswe[key]/max(hist_dswe[key]), 'b-')
        plt.plot(0.5 * (bins[0:-1] + bins[1:]), hist_dswe_thresh[key]/max(hist_dswe_thresh[key]),  'r-')

        plt.xlim((-41, 71))
        plt.xticks([-40, -20, 0, 20, 40, 60])
        ylim = plt.gca().get_ylim()
        plt.gca().axes.yaxis.set_ticks(ylim)
        plt.gca().set_yticklabels(['0', 'max'])
        if ii == 2:
            plt.xlabel(r'Temporal mean 24-day $\Delta$SWE [mm]')
        else:
            plt.gca().set_xticklabels([])
            plt.ylabel('')
    plt.tight_layout()

    plt.savefig(fig_dir+'seasonal_distributions_b.png')

    for ii in np.arange(len(plot_keys)):
        key = plot_keys[ii]
        print('dswe mean [mm], s.d. [mm] for ', key, ' = ', np.nanmean(mean_dswe[key])*1000., np.nanstd(mean_dswe[key])*1000.)

    for ii in np.arange(len(plot_keys)):
        keyii = plot_keys[ii]
        good_mask = np.reshape(good_mask, par.dim)
        for jj in np.arange(ii+1, len(plot_keys)):
            keyjj = plot_keys[jj]
            sii = mean_dswe[keyii][good_mask]
            sjj = mean_dswe[keyjj][good_mask]
            sii-=np.mean(sii)
            sjj-=np.mean(sjj)
            #corr = np.nanmean(mean_dswe[keyii]*mean_dswe[keyjj])/(np.nanmean(mean_dswe[keyii]**2) * np.nanmean(mean_dswe[keyjj]**2)) ** 0.5
            corr = np.mean(sii * sjj) / (np.mean(sii ** 2) * np.mean(sjj ** 2)) ** 0.5
            print('corr (', keyii, keyjj, ') = ', corr)
    plt.show()

#make figs for publication
def make_figs(cfg):
    rdcras_to_geotiff(cfg['stack']._dir + 'rmli_3_12/water_mask_3_12.ras', outtif = fig_dir+'water_mask_RS2SLA27D_3_12.geo.tif', stack=cfg['stack'], looks_str=cfg['looks_str'], interp_mode=0)
    rdcras_to_geotiff(cfg['stack']._dir + 'rmli_3_12/rmli_3_12.ave.ras', outtif = fig_dir+'rmli_ave_RS2SLA27D_3_12.geo.tif', stack=cfg['stack'], looks_str=cfg['looks_str'])
    rdcras_to_geotiff(cfg['stack']._dir + 'dem_3_12/topo/dphi_dswe.rdc.png', outtif = fig_dir+'xi_RS2SLA27D_3_12.geo.tif', stack=cfg['stack'], looks_str=cfg['looks_str'])
    rdcras_to_geotiff(cfg['stack']._dir + 'base_adjust_all/20120317_20120410.diff_init.ras', outtif = fig_dir+'20120317_20120410.diff_init.geo.tif', stack=cfg['stack'], looks_str='11_44', interp_mode=0)
    estimation_drill_down_plot(cfg)
    effect_of_wrong_density()
    inuvik_climate_normals_plot()
    sim_multiplot(cfg, [49], 'bare', plot_singles=True)
    sim_multiplot(cfg, [49], 'nalcms', plot_singles=True)
    sim_multiplot(cfg, np.arange(132), 'bare', plot_history=True, cache_filename = '/local-scratch/users/jaysone/projects_active/inuvik/RS2_SLA27_D/full_scene/snowmodel_3_12/bare/err_all.pickle')
    sim_multiplot(cfg, np.arange(132), 'nalcms', plot_history=True, cache_filename = '/local-scratch/users/jaysone/projects_active/inuvik/RS2_SLA27_D/full_scene/snowmodel_3_12/nalcms/err_all.pickle')
    other_phase_sources(cfg)
    monte_carlo(cfg, pair_name='20120317_20120410', plot=True)
    insitu_scatterplot(cfg)
    temporal_validation_plot(cfg)
    swe_maps_all(cfg)
    blown_snow_analysis()
    delta_k_comparison(cfg)
    seasonal_mean_maps(cfg)
    seasonal_mean_maps(cfg, thresh_xi_sd=(7.585708141326904, 1e6))

if __name__ == '__main__':
    cfg = cfg_default
    make_figs(cfg)
    plt.show()

