import datetime
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates
import matplotlib.cm
from scipy import special
import .era5 as era5

def get_era5_snow_study_series(loc, start_date, end_date, coastal=False):
    codes = ['rsn', 'sd', '10u', '10v', 'z', '2t', '2d', 'tsn', 'stl1', 'es', 'smlt', 'sf', 'tp', 'ro']
    return_dict = era5.get_dataset(loc, start_date, end_date, codes, coastal=coastal)
    return_dict.update(era5.get_derived_codes(return_dict))
    #add derived parameters
    return_dict.update(_derived_snow_study_series(return_dict))
    return return_dict

snow_param_full_names = era5.param_full_names
snow_param_full_names.update({
    'rh': '2m relative humidity',
    '10m': '10 m wind magnitude',
    '10d': '10 m wind direction',
    'rsn_new': 'new snow density - derived',
    'sd_snow': 'homogenous snow depth - derived'
    })

snow_param_units = era5.param_units
snow_param_units.update({
    'rh': '%',
    '10m': 'm/s',
    '10d': 'degrees',
    'rsn_new': 'kg/m^3',
    'sd_snow': 'm'
    })

for param in snow_param_units:
    if snow_param_units[param] == 'K':
        snow_param_units[param] = 'deg. C'

def _derived_snow_study_series(dict):
    #adds addition computed parameters to the dictionary
    nt = len(dict['2t'])
    dict_new = {}
    #true snow depth
    dict_new['sd_snow'] = dict['sd']/(dict['rsn']/1000.0) #m

    #density of fresh snow
    rsn_new  = np.zeros(nt)
    del_sd = (dict['sd'][1:] - dict['sd'][0:-1])
    del_sd_snow = (dict_new['sd_snow'][1:] - dict_new['sd_snow'][0:-1])
    mask = (dict_new['sd_snow'][1:] - dict_new['sd_snow'][0:-1]) > 0.0001
    rsn_new[1:][mask] = del_sd[mask]/del_sd_snow[mask]*1000.0
    dict_new['rsn_new'] = rsn_new

    #relative humidity
    def es(t):
        #implements the  August-Roche-Magnus formula for saturation vapour pressure
        #t: temperature in Celcius
        #see Lawrence, Mark.  American Metr. Soc 2005 eq 6
        return 6.1094*np.exp(17.625*t/(t+243.04))

    dict_new['rh'] = 100*es(dict['2d'])/es(dict['2t'])

    #snow seasons
    snow_mask = dict['sd'] > 0.002
    seasonal_snow_mask = np.ones(nt)
    seed_idx_list = [0, nt-1] #assume beginning and end are in summer
    start_time = datetime.datetime.strptime(dict['start_date'], '%Y%m%d')

    for ii in np.arange(0,nt):
        time_ii = start_time+datetime.timedelta(dict['dt']*ii)
        if time_ii.month == 7 and time_ii.day == 1 and time_ii.hour == 0:
            seed_idx_list.append(ii)
    for seed_idx in seed_idx_list:
        idx = seed_idx
        while idx < nt and snow_mask[idx] == 0:
            seasonal_snow_mask[idx] = 0
            idx +=1
        idx = seed_idx
        while idx > -1 and snow_mask[idx] != 0:
            seasonal_snow_mask[idx] = 0
            idx -= 1

    dict_new['seasonal_snow_mask']=seasonal_snow_mask
    return dict_new


def plot_snow_study_series(dict):
    #group similar codes into plots
    snowfall_codes = ['sf', 'rsn_new']
    other_snow_codes = ['sd', 'sd_snow', 'rsn', 'es', 'smlt']
    temperature_codes = ['2t', 'tsn', 'stl1']
    wind_codes = ['10m', '10d']
    main_params = ['2t', '10m', '10d', 'sf', 'es', 'smlt', 'sd']
    fels_codes = ['2t', 'sd', 'smlt', 'tp', 'ro']
    params_1 = ['2t', '10m', '10d']
    params_2 = ['sf', 'es', 'smlt', 'sd']
    blowing_snow_params = ['2t', '10m', '10d', 'sf', 'bsi']

    for codes in [temperature_codes, snowfall_codes, other_snow_codes, wind_codes, params_1, params_2, blowing_snow_params]:
        fig, axarr = plt.subplots(len(codes), sharex=True)
        for ii, code in enumerate(codes):
            axarr[ii].plot(dict['times'], dict[code])
            if code in snow_param_full_names:
                axarr[ii].set_title(snow_param_full_names[code])
            else:
                axarr[ii].set_title(code)
            if code in era5.param_units:
                axarr[ii].set_ylabel(era5.param_units[code])
            axarr[ii].format_xdata = matplotlib.dates.DateFormatter('%Y-%m-%d')

        fig.autofmt_xdate()

    #wind distribution polar plot
    snow_mask = dict['sd'] > 0.001
    abins = np.linspace(0, 2 * np.pi, 36)  # 0 to 360 in steps of 360/N.
    sbins = np.linspace(0, 15, 30)
    H, xedges, yedges = np.histogram2d(dict['10d'][snow_mask]*np.pi/180, dict['10m'][snow_mask], bins=(abins, sbins))
    H = H*dict['dt']*24.0 #convert to cumulative hours per bin
    theta, r = np.mgrid[0:2 * np.pi:36j, 0:15:30j]
    fig, ax = plt.subplots(subplot_kw={'projection':'polar'})
    ax.set_facecolor('black')
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    p1 = ax.pcolormesh(theta, r, H)
    plt.colorbar(p1, ax=ax)
    ax.tick_params(axis='y', colors='white')


def li_and_pomeroy(U, T, dt_hours):
    #from Li and Pomeroy, 1997
    #I = np.log(np.arange(6, 148, 6))
    I = np.log(dt_hours)
    U_bar = 0.365*T + 0.00706*(T**2) + 0.9*I +11.2
    delta = 0.145*T + 0.00196*(T**2) + 4.3
    z = (U-U_bar)/delta
    if np.isscalar(z):
        P = 0.5+0.5*special.erf(z/(2**0.5))
    else:
        P = [0.5+0.5*special.erf(z_elem/(2**0.5)) for z_elem in z]
    return P

def blown_snow_index2(dict):
    ns = len(dict['2t'])
    bsi = np.zeros(ns)
    for ii in np.arange(1, ns):
        bsi[ii] = li_and_pomeroy(dict['10m'][ii], dict['2t'][ii], 1)
        if dict['sd'][ii] < 0.001:
            bsi[ii] = 0.0
        if dict['10m'][ii]<3.0:
            bsi[ii] = 0.0
    return bsi

def get_era5_pairwise_data(dict, acq_dates, acq_fractional_day=0, plot=True):
    len_dict = len(dict['times'])
    start_time = datetime.datetime.strptime(dict['start_date'], '%Y%m%d')
    acq_times = [datetime.datetime.strptime(date, '%Y%m%d')+ datetime.timedelta(acq_fractional_day) for date in acq_dates]
    acq_idx = [np.rint((acq_time - start_time).days/dict['dt']).astype(np.int) for acq_time in acq_times]
    nd = len(acq_dates)
    max_sd = np.zeros((nd, nd)) + np.inf
    min_sd = np.zeros((nd, nd))+ np.inf
    max_2t = np.zeros((nd, nd))+ np.inf
    delta_sd = np.zeros((nd, nd))+ np.inf
    cum_sm = np.zeros((nd,nd))+ np.inf
    cum_bsi = np.zeros((nd,nd))+ np.inf
    cum_bsi_vec = np.zeros((nd, nd), dtype=complex) + np.inf
    snow_seasons = np.zeros((nd, nd))+ np.inf
    mask = np.zeros((nd,nd))
    for ii in range(nd):
        for jj in range(ii,nd):
            if (acq_idx[ii] < len_dict) and  (acq_idx[jj] < len_dict) and acq_idx[ii]>=0 and acq_idx[jj] >=0:
                max_sd[ii, jj] = max(dict['sd'][acq_idx[ii]], dict['sd'][acq_idx[jj]])
                max_sd[jj, ii] = max_sd[ii, jj]
                min_sd[ii, jj] = min(dict['sd'][acq_idx[ii]], dict['sd'][acq_idx[jj]])
                min_sd[jj, ii] = min_sd[ii, jj]
                max_2t[ii, jj] = np.max(dict['2t'][acq_idx[ii]:acq_idx[jj]+1])
                max_2t[jj, ii] = max_2t[ii, jj]
                delta_sd[ii, jj] =dict['sd'][acq_idx[jj]] - dict['sd'][acq_idx[ii]]

                delta_sd[jj, ii] = -1*delta_sd[ii, jj]
                cum_sm[ii, jj] = np.sum(dict['smlt'][acq_idx[ii]:acq_idx[jj]+1])
                cum_sm[jj, ii] = -1*cum_sm[ii, jj]
                cum_bsi[ii, jj] = np.sum(dict['bsi'][acq_idx[ii]:acq_idx[jj]+1])
                cum_bsi[jj, ii] = 0#-1* cum_bsi[ii, jj]
                cum_bsi_vec[ii, jj] = np.sum(dict['bsi'][acq_idx[ii]:acq_idx[jj]+1]*np.exp(1j*dict['10d'][acq_idx[ii]:acq_idx[jj]+1]/180*np.pi))
                cum_bsi_vec[jj, ii] = 0

                snow_seasons[ii, jj] = np.prod(dict['seasonal_snow_mask'][acq_idx[ii]:acq_idx[jj]+1])
                snow_seasons[jj, ii] = snow_seasons[ii, jj]
                mask[ii, jj] = 1
            #else: uses data not present in dict results so just skip
            if ii == jj:
                mask[ii, jj] = 0

    snow_ifg_mask = (min_sd > 0.001) * (cum_sm < 0.001) * mask

    era5_pairwise = {}
    era5_pairwise['mask'] = mask
    era5_pairwise['max_sd'] = max_sd
    era5_pairwise['min_sd'] = min_sd
    era5_pairwise['max_2t'] = max_2t
    era5_pairwise['delta_sd'] = delta_sd
    era5_pairwise['cum_sm'] = cum_sm
    era5_pairwise['cum_bsi'] = cum_bsi
    era5_pairwise['cum_bsi_vec_mag'] = np.abs(cum_bsi_vec)
    era5_pairwise['cum_bsi_vec_dir'] = np.angle(cum_bsi_vec)*180/np.pi
    era5_pairwise['snow_seasons'] = snow_seasons

    #network selection keys
    era5_pairwise['snow_free'] = max_sd < 0.001
    era5_pairwise['snow_free'][mask == 0] = False
    era5_pairwise['dry_snow_intra'] = np.logical_and(min_sd > 0.001, cum_sm < 0.001)
    era5_pairwise['dry_snow_intra'][mask == 0] = False

    if plot:
        #all
        plt.figure()
        for ii, key in enumerate(era5_pairwise):
            masked_array = np.ma.array(era5_pairwise[key], mask=(era5_pairwise['mask']==0))
            cmap = matplotlib.cm.jet
            cmap.set_bad('black', 1.)
            plt.subplot(3, len(era5_pairwise)//3+1, ii+1)
            plt.imshow(masked_array.T, interpolation='nearest', cmap=cmap)
            plt.title(key)

        plot_list = ['cum_sm', 'delta_sd', 'snow_seasons', 'min_sd', 'cum_bsi', 'cum_bsi_vec_mag', 'cum_bsi_vec_dir', 'snow_free', 'dry_snow_intra']
        plt.figure()
        for ii, key in enumerate(plot_list):
            masked_array = np.ma.array(era5_pairwise[key], mask=(era5_pairwise['mask']==0))
            cmap = matplotlib.cm.jet
            cmap.set_bad('black', 1.)
            plt.subplot(2, len(plot_list)//2+1, ii+1)
            plt.imshow(masked_array.T, interpolation='nearest', cmap=cmap)
            plt.title(key)

        plt.figure()
        for ii, key in enumerate(plot_list):
            masked_array = np.ma.array(era5_pairwise[key], mask=(snow_ifg_mask==0))
            cmap = matplotlib.cm.jet
            cmap.set_bad('black', 1.)
            plt.subplot(2, len(plot_list)//2+1, ii+1)
            plt.imshow(masked_array.T, interpolation='nearest', cmap=cmap)
            plt.title(key)
            #plt.colorbar()

    return era5_pairwise